"""Implements the MDCEV model as a generic class

Michel Bierlaire
Sun Apr 7 16:52:33 2024
"""

from __future__ import annotations

import logging
import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from functools import lru_cache
from typing import Any, Iterable

import numpy as np
import pandas as pd
from scipy.optimize import minimize

from biogeme.biogeme import BIOGEME
from biogeme.calculator import (
    evaluate_expression,
    get_value_and_derivatives,
)
from biogeme.database import Database
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Beta, Elem, Expression, Numeric, bioMultSum, exp, log
from biogeme.function_output import FunctionOutput
from biogeme.results_processing import EstimationResults
from biogeme.tools.checks import validate_dict_types

logger = logging.getLogger(__name__)

SMALLEST_NON_ZERO_NUMBER = 1.0e-6


@dataclass(frozen=True)
class MdcevConfiguration:
    """Identify various configurations of the model"""

    gamma_is_none: bool
    price_is_none: bool


class Mdcev(ABC):

    def __init__(
        self,
        model_name: str,
        baseline_utilities: dict[int, Expression],
        gamma_parameters: dict[int, Expression | None],
        alpha_parameters: dict[int, Expression] | None = None,
        scale_parameter: Expression | None = None,
        weights: Expression | None = None,
    ) -> None:
        """ """

        self.model_name: str = model_name
        self.baseline_utilities: dict[int, Expression] = baseline_utilities
        self.gamma_parameters: dict[int, Expression | None] = gamma_parameters
        self.alpha_parameters: dict[int, Expression] | None = alpha_parameters
        self.scale_parameter: Expression | None = scale_parameter
        self.weights: Expression | None = weights
        self._estimation_results: EstimationResults | None = None
        self.database: Database | None = None
        # Check the numbering
        self.alternatives: set[int] = set(self.baseline_utilities)
        # Map the indices with the keys
        self.index_to_key: list[int] = [key for key in self.alternatives]
        self.key_to_index: dict[int, int] = {
            key: index for index, key in enumerate(self.index_to_key)
        }
        if len(self.alternatives) != len(list(self.baseline_utilities)):
            error_msg = (
                f'Some alternatives appear more than once in baseline utilities: '
                f'{list(self.baseline_utilities)}'
            )
            raise BiogemeError(error_msg)
        if error_messages := self._verify_dict_keys_against_set(
            dict_to_check=self.gamma_parameters
        ):
            error_msg = f'Gamma parameters: {", ".join(error_messages)}'
            raise BiogemeError(error_msg)

        if self.alpha_parameters:
            if error_messages := self._verify_dict_keys_against_set(
                dict_to_check=self.alpha_parameters
            ):
                error_msg = f'Alpha parameters: {", ".join(error_messages)}'
                raise BiogemeError(error_msg)

        # Check if there is an outside good. It correspond to a gamma parameter set to None.
        none_keys = [
            key for key, value in self.gamma_parameters.items() if value is None
        ]
        if len(none_keys) > 1:
            error_msg = f'Only one outside good is allowed, not {len(none_keys)}'
            raise BiogemeError(error_msg)
        self.outside_good_key = none_keys[0] if len(none_keys) != 0 else None

        validate_dict_types(self.baseline_utilities, 'baseline_utilities', Expression)
        validate_dict_types(self.gamma_parameters, 'gamma_parameters', Expression)
        if self.alpha_parameters is not None:
            validate_dict_types(self.alpha_parameters, 'alpha_parameters', Expression)
        if self.scale_parameter is not None:
            if not isinstance(self.scale_parameter, Expression):
                error_msg = f'Expecting a Biogeme expression and not {type(self.scale_parameter)}'
                raise BiogemeError(error_msg)

    @property
    def outside_good_index(self) -> int | None:
        """Obtain  the index of the outside good."""
        if self.outside_good_key is None:
            return None
        return self.key_to_index[self.outside_good_key]

    @property
    def estimation_results(self) -> EstimationResults | None:
        """Property for the estimation results"""
        return self._estimation_results

    @estimation_results.setter
    def estimation_results(self, the_results: EstimationResults):
        self._estimation_results = the_results
        self._update_parameters_in_expressions()

    @property
    def number_of_alternatives(self) -> int:
        return len(self.alternatives)

    def _verify_dict_keys_against_set(self, dict_to_check: dict[int, Any]) -> list[str]:
        # Convert dictionary keys to a set
        dict_keys = set(dict_to_check.keys())

        # Find keys in the dictionary that are not in the reference set
        extra_keys = dict_keys - self.alternatives

        # Find keys in the reference set that are not in the dictionary
        missing_keys = self.alternatives - dict_keys

        # Constructing the error message based on the findings
        error_messages = []

        if extra_keys:
            error_messages.append(f'Extra alternatives in dictionary: {extra_keys}.')

        if missing_keys:
            error_messages.append(
                f'Missing alternatives from dictionary: {missing_keys}.'
            )
        return error_messages

    @lru_cache
    def calculate_baseline_utility(
        self, alternative_id: int, one_observation: Database
    ) -> float:
        """As this function may be called many times with the same input in forecasting mode, we use the
        lru_cache decorator."""
        assert one_observation.num_rows() == 1
        if self.estimation_results:
            return evaluate_expression(
                expression=self.baseline_utilities[alternative_id],
                numerically_safe=False,
                database=one_observation,
                betas=self.estimation_results.get_beta_values(),
                aggregation=True,
            )

        return evaluate_expression(
            expression=self.baseline_utilities[alternative_id],
            numerically_safe=False,
            database=one_observation,
            aggregation=True,
        )

    @abstractmethod
    def transformed_utility(
        self,
        the_id: int,
        the_consumption: Expression,
    ) -> Expression:
        """Calculates the transformed utility for one alternative. Used for estimation"""
        pass

    def calculate_utilities(
        self, consumption: dict[int, Expression]
    ) -> dict[int, Expression]:

        return {
            alt_id: self.transformed_utility(alt_id, the_consumption)
            for alt_id, the_consumption in consumption.items()
        }

    @abstractmethod
    def calculate_log_determinant_one_alternative(
        self, the_id: int, consumption: Expression
    ) -> Expression:
        pass

    def calculate_log_determinant_entries(
        self, consumption: dict[int, Expression]
    ) -> dict[int, Expression]:
        return {
            alt_id: self.calculate_log_determinant_one_alternative(
                alt_id, the_consumption
            )
            for alt_id, the_consumption in consumption.items()
        }

    @abstractmethod
    def calculate_inverse_of_determinant_one_alternative(
        self, the_id: int, consumption: Expression
    ) -> Expression:
        pass

    def calculate_inverse_of_determinant_entries(
        self, consumption: dict[int, Expression]
    ) -> dict[int, Expression]:
        return {
            alt_id: self.calculate_inverse_of_determinant_one_alternative(
                alt_id, the_consumption
            )
            for alt_id, the_consumption in consumption.items()
        }

    def info_gamma_parameters(self) -> str:
        """Provides logging information about the outside good"""
        none_count = sum(1 for value in self.gamma_parameters.values() if value is None)
        if none_count == 0:
            report = 'No outside good is included in the model.'
            logger.info(report)
            return report
        if none_count == 1:
            report = 'One outside good is included in the model.'
            logger.info(report)
            return report

        report = (
            'Several outside goods are included in the model. If it is intentional, ignore this warning. '
            'If not, update the definition of the gamma parameters. The outside good, if any, must be associated with '
            'None, and all other alternative with a gamma parameter.'
        )
        logger.warning(report)
        return report

    def loglikelihood(
        self,
        number_of_chosen_alternatives: Expression,
        consumed_quantities: dict[int, Expression],
    ) -> Expression:
        """Generate the Biogeme formula for the log probability of the MDCEV model

        :param number_of_chosen_alternatives: see the module documentation :mod:`biogeme.mdcev_no_outside_good`
        :param consumed_quantities: see the module documentation :mod:`biogeme.mdcev_no_outside_good`

        A detailed explanation is provided in the technical report
        "Estimating the MDCEV model with Biogeme"

        """
        validate_dict_types(consumed_quantities, 'consumed_quantities', Expression)

        utilities = self.calculate_utilities(consumption=consumed_quantities)
        log_determinant_entries = self.calculate_log_determinant_entries(
            consumption=consumed_quantities
        )
        inverse_of_determinant_entries = self.calculate_inverse_of_determinant_entries(
            consumption=consumed_quantities
        )

        # utility of chosen goods
        utility_terms = [
            Elem({0: 0.0, 1: util}, consumed_quantities[i] > 0)
            for i, util in utilities.items()
        ]
        if self.scale_parameter is None:
            baseline_term = bioMultSum(utility_terms)
        else:
            baseline_term = self.scale_parameter * bioMultSum(utility_terms)

        # Determinant: first term
        first_determinant_terms = [
            Elem({0: 0.0, 1: z}, consumed_quantities[i] > 0)
            for i, z in log_determinant_entries.items()
        ]
        first_determinant = bioMultSum(first_determinant_terms)

        # Determinant: second term
        second_determinant_terms = [
            Elem({0: 0.0, 1: z}, consumed_quantities[i] > 0)
            for i, z in inverse_of_determinant_entries.items()
        ]
        second_determinant = log(bioMultSum(second_determinant_terms))

        # Logsum
        if self.scale_parameter is None:
            terms_for_logsum = [exp(util) for util in utilities.values()]
        else:
            terms_for_logsum = [
                exp(self.scale_parameter * util) for util in utilities.values()
            ]
        logsum_term = number_of_chosen_alternatives * log(bioMultSum(terms_for_logsum))

        log_prob = baseline_term + first_determinant + second_determinant - logsum_term
        # Scale parameter
        if self.scale_parameter is not None:
            log_prob += (number_of_chosen_alternatives - 1) * log(self.scale_parameter)

        return log_prob

    def estimate_parameters(
        self,
        database: Database,
        number_of_chosen_alternatives: Expression,
        consumed_quantities: dict[int, Expression],
        **kwargs,
    ) -> EstimationResults:
        """Generate the Biogeme formula for the log probability of the MDCEV model

        :param database: data needed for the estimation of the parameters, in Biogeme format.
        :param number_of_chosen_alternatives: see the module documentation :mod:`biogeme.mdcev_no_outside_good`
        :param consumed_quantities: see the module documentation :mod:`biogeme.mdcev_no_outside_good`
        :param **kwargs: additional parameters that are transmitted as such to the constructor of the Biogeme object.


        A detailed explanation is provided in the technical report
        "Estimating the MDCEV model with Biogeme":
        """
        logprob = self.loglikelihood(
            number_of_chosen_alternatives=number_of_chosen_alternatives,
            consumed_quantities=consumed_quantities,
        )

        # Create the Biogeme object
        if self.weights is None:
            the_biogeme = BIOGEME(database, logprob, **kwargs)
        else:
            formulas = {'log_like': logprob, 'weight': self.weights}
            the_biogeme = BIOGEME(database, formulas, **kwargs)
        the_biogeme.modelName = self.model_name
        self.estimation_results = the_biogeme.estimate()
        return self.estimation_results

    @abstractmethod
    def _list_of_expressions(self) -> list[Expression]:
        """Extract the list of expressions involved in the model"""
        raise NotImplementedError

    def _update_parameters_in_expressions(self) -> None:
        """Update the value of the unknown parameters in expression, after estimation"""
        if self._estimation_results is None:
            error_msg = 'No estimation result is available'
            raise BiogemeError(error_msg)

        betas = self._estimation_results.get_beta_values()
        all_expressions = [
            expression
            for group in [
                self.baseline_utilities.values(),
                self.gamma_parameters.values(),
                (
                    self.alpha_parameters.values()
                    if self.alpha_parameters is not None
                    else []
                ),
                [self.scale_parameter] if self.scale_parameter is not None else [],
                [self.weights] if self.weights is not None else [],
            ]
            for expression in group
            if expression is not None
        ]

        # ADd the expressions from the child class
        all_expressions += self._list_of_expressions()

        # Update each expression with the new beta values
        for expression in all_expressions:
            expression.change_init_values(betas)

    @abstractmethod
    def utility_one_alternative(
        self,
        the_id: int,
        the_consumption: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Utility needed for forecasting"""
        raise NotImplementedError

    @abstractmethod
    def utility_expression_one_alternative(
        self,
        the_id: int,
        the_consumption: Expression,
        unscaled_epsilon: Expression,
    ) -> Expression:
        """Utility expression. Used only for code validation."""
        raise NotImplementedError

    def sum_of_utilities(
        self, consumptions: np.ndarray, epsilon: np.ndarray, data_row: Database
    ) -> float:
        """Calculates the sum of all utilities. Used for forecasting."""
        try:
            utilities = [
                self.utility_one_alternative(
                    the_id=key,
                    the_consumption=float(consumptions[index]),
                    epsilon=float(epsilon[index]),
                    one_observation=data_row,
                )
                for index, key in enumerate(self.index_to_key)
            ]
        except IndexError as e:
            raise e

        return np.sum(utilities)

    def forecast_bruteforce_one_draw(
        self, one_row_database: Database, total_budget: float, epsilon: np.ndarray
    ) -> dict[int, float] | None:
        """Forecast the optimal expenditures given a budget and one realization of epsilon for each alternative.
        If there is an issue with the optimization algorithm, None is returned.
        """

        if len(epsilon) != self.number_of_alternatives:
            error_msg = f'epsilon must be a vector of size {self.number_of_alternatives}, not {epsilon.shape}'
            raise BiogemeError(error_msg)

        if len(one_row_database.data) != 1:
            error_msg = f'Expecting exactly one row, not {len(one_row_database.data)}'
            raise BiogemeError(error_msg)

        self.database = one_row_database

        # If there is an outside good, we need to impose that the corresponding consumption is not zero.
        # To do that, we impose that it is equal to the exponential of a new variable.

        outside_good = self.outside_good_key is not None

        def objective_function(x: np.ndarray) -> float:
            """Objective function. The number of variables is the number of
            alternatives"""
            result = self.sum_of_utilities(
                consumptions=x, epsilon=epsilon, data_row=one_row_database
            )
            return -result

        def budget_constraint(x: np.ndarray) -> float:
            """Budget constraint when there is no outside good. The number of variables is the number of
            alternatives"""
            return total_budget - x.sum()

        def opposite_budget_constraint(x: np.ndarray) -> float:
            """Budget constraint when there is no outside good. The number of variables is the number of
            alternatives"""
            return x.sum() - total_budget

        constraints = [
            {
                'type': 'ineq',
                'fun': budget_constraint,
            },
            {
                'type': 'ineq',
                'fun': opposite_budget_constraint,
            },
        ]

        number_of_alternatives = len(self.alternatives)

        # Bounds
        bounds: list[tuple[float | None, float | None]] = [
            (0, total_budget) for _ in range(number_of_alternatives)
        ]

        if self.outside_good_index is not None:
            bounds[self.outside_good_index] = (SMALLEST_NON_ZERO_NUMBER, total_budget)

        # Starting point
        # We split the budget equally across alternatives
        initial_consumption = total_budget / number_of_alternatives
        # We split equally across alternatives
        initial_guess = np.array([initial_consumption] * number_of_alternatives)

        # There is a bug in the minimize function. It generates a warning that crashes the script.
        # We are converting this warning into an exception in order to catch it.
        # Convert specific warnings to exceptions
        warnings.filterwarnings(
            'error',
            message='Values in x were outside bounds during a minimize step, clipping to bounds',
        )

        try:
            optimization_result = minimize(
                objective_function,
                initial_guess,
                method='SLSQP',
                bounds=bounds,
                constraints=constraints,
            )
            optimal_consumption = {
                self.index_to_key[index]: optimization_result.x[index]
                for index in range(number_of_alternatives)
            }
            for alternative in self.alternatives:
                if alternative not in optimal_consumption:
                    optimal_consumption[alternative] = 0
            return optimal_consumption
        except RuntimeWarning as e:
            logger.warning(e)
            return None

    @abstractmethod
    def derivative_utility_one_alternative(
        self,
        the_id: int,
        the_consumption: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Used in the optimization problem solved for forecasting tp calculate the dual variable."""
        raise NotImplementedError

    def optimal_consumption(
        self,
        chosen_alternatives: Iterable[int],
        dual_variable: float,
        epsilon: np.ndarray,
        one_observation: Database,
    ) -> dict[int, float]:
        """Analytical calculation of the optimal consumption if the dual variable is known."""
        try:
            result = {
                alt_id: self.optimal_consumption_one_alternative(
                    the_id=alt_id,
                    dual_variable=dual_variable,
                    epsilon=float(epsilon[self.key_to_index[alt_id]]),
                    one_observation=one_observation,
                )
                for alt_id in chosen_alternatives
            }
        except KeyError as e:
            raise e
        return result

    @abstractmethod
    def optimal_consumption_one_alternative(
        self,
        the_id: int,
        dual_variable: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Analytical calculation of the optimal consumption if the dual variable is known."""
        raise NotImplementedError

    @abstractmethod
    def lower_bound_dual_variable(
        self,
        chosen_alternatives: set[int],
        one_observation: Database,
        epsilon: np.ndarray,
    ) -> float:
        """Method providing model specific bounds on the dual variable. It not overloaded,
        default values are used.

        :param chosen_alternatives: list of alternatives that are chosen at the optimal solution
        :param one_observation: data for one observation.
        :param epsilon: draws from the error term.
        :return: a lower bound on the dual variable, such that the expenditure calculated for any larger value is
        well-defined and non negative.
        """
        raise NotImplementedError

    def is_next_alternative_chosen(
        self,
        chosen_alternatives: set[int],
        candidate_alternative: int,
        derivative_zero_expenditure_candidate: float,
        database: Database,
        total_budget: float,
        epsilon: np.ndarray,
    ) -> tuple[bool, float | None]:
        # If the candidate alternative is the outside good, it is in the choice set by definition.
        model_lower_bound = self.lower_bound_dual_variable(
            chosen_alternatives=chosen_alternatives,
            one_observation=database,
            epsilon=epsilon,
        )
        if derivative_zero_expenditure_candidate < model_lower_bound:
            # Lower bound is violated. Alternative not in choice set.
            return False, model_lower_bound

        candidate_set = chosen_alternatives | {candidate_alternative}
        # Derive the optimal consumption
        estimate_optimal_consumption_candidate = self.optimal_consumption(
            chosen_alternatives=candidate_set,
            dual_variable=derivative_zero_expenditure_candidate,
            epsilon=epsilon,
            one_observation=database,
        )
        total_consumption_candidate = sum(
            estimate_optimal_consumption_candidate.values()
        )

        if total_budget <= total_consumption_candidate:
            # Total budget is overpassed. Alternative not in the choice set.
            return False, derivative_zero_expenditure_candidate
        return True, None

    def identification_chosen_alternatives(
        self,
        database: Database,
        total_budget: float,
        epsilon: np.ndarray,
    ) -> tuple[set[int], float, float]:
        """Algorithm identifying the chosen alternatives at the optimal solution

        :param database: one row of the database.
        :param total_budget: total budget for the constraint.
        :param epsilon: vector of draws from the error term.
        :return: the set of chosen alternatives, as well as a lower and an upper bound on the dual variable.
        """
        the_unsorted_w: dict[int, float] = {
            alt_id: self.derivative_utility_one_alternative(
                the_id=alt_id,
                one_observation=database,
                the_consumption=0,
                epsilon=float(epsilon[self.key_to_index[alt_id]]),
            )
            for alt_id in self.alternatives
            if alt_id != self.outside_good_key
        }

        # Sort the dictionary by values (descending order)

        ordered_alternatives: list[int] = [
            alt_id
            for alt_id, _ in sorted(
                the_unsorted_w.items(), key=lambda item: item[1], reverse=True
            )
        ]

        chosen_alternatives: set[int] = (
            set() if self.outside_good_key is None else {self.outside_good_key}
        )

        last_chosen_alternative = None
        for candidate_alternative_id in ordered_alternatives:
            derivative_zero_expenditure_candidate = the_unsorted_w[
                candidate_alternative_id
            ]
            next_chosen, lower_bound = self.is_next_alternative_chosen(
                chosen_alternatives=chosen_alternatives,
                candidate_alternative=candidate_alternative_id,
                derivative_zero_expenditure_candidate=derivative_zero_expenditure_candidate,
                database=database,
                total_budget=total_budget,
                epsilon=epsilon,
            )
            if not next_chosen:
                upper_bound = (
                    np.finfo(np.float64).max
                    if last_chosen_alternative is None
                    else the_unsorted_w[last_chosen_alternative]
                )
                return (
                    chosen_alternatives,
                    lower_bound,
                    upper_bound,
                )

            chosen_alternatives |= {candidate_alternative_id}
            last_chosen_alternative = candidate_alternative_id
        # The full choice set is chosen
        lower_bound = self.lower_bound_dual_variable(
            chosen_alternatives=chosen_alternatives,
            one_observation=database,
            epsilon=epsilon,
        )
        upper_bound = the_unsorted_w[ordered_alternatives[-1]]
        return chosen_alternatives, lower_bound, upper_bound

    def forecast_bisection_one_draw(
        self,
        one_row_of_database: Database,
        total_budget: float,
        epsilon: np.ndarray,
        tolerance_dual=1.0e-13,
        tolerance_budget=1.0e-13,
    ) -> dict[int, float]:
        chosen_alternatives, lower_bound, upper_bound = (
            self.identification_chosen_alternatives(
                database=one_row_of_database, total_budget=total_budget, epsilon=epsilon
            )
        )

        if lower_bound > upper_bound:
            error_msg = (
                f'Lower bound: {lower_bound} larger than upper bound: {upper_bound}. '
                f'Chosen alternatives: {chosen_alternatives}'
            )
            raise BiogemeError(error_msg)

        # Estimate the dual variable by bisection
        total_consumption = 0
        continue_iterations = True
        for _ in range(5000):
            if not continue_iterations:
                break
            dual_variable = (lower_bound + upper_bound) / 2
            optimal_consumption = self.optimal_consumption(
                chosen_alternatives=chosen_alternatives,
                dual_variable=dual_variable,
                epsilon=epsilon,
                one_observation=one_row_of_database,
            )
            negative_consumption = any(
                value < 0 for value in optimal_consumption.values()
            )
            if negative_consumption:
                raise ValueError('Negative consumption')

            if negative_consumption:
                # If any consumption is negative, we update the upper bound
                upper_bound = dual_variable
            else:
                total_consumption = sum(optimal_consumption.values())
                if total_consumption < total_budget:
                    upper_bound = dual_variable

                elif total_consumption > total_budget:
                    lower_bound = dual_variable
                # Stopping criteria
                if (upper_bound - lower_bound) <= tolerance_dual:
                    continue_iterations = False
                elif np.abs(total_consumption - total_budget) <= tolerance_budget:
                    continue_iterations = False

        dual_variable = (lower_bound + upper_bound) / 2
        optimal_consumption = self.optimal_consumption(
            chosen_alternatives=chosen_alternatives,
            dual_variable=dual_variable,
            epsilon=epsilon,
            one_observation=one_row_of_database,
        )
        for alternative in self.alternatives:
            if alternative not in optimal_consumption:
                optimal_consumption[alternative] = 0
        return optimal_consumption

    def validation_one_alternative_one_consumption(
        self, alternative_id: int, value_consumption: float, one_row: Database
    ) -> list[str]:
        """Validation of some of the abstract methods."""
        error_messages = []

        value_epsilon = 0.01
        consumption = Beta('consumption', value_consumption, None, None, 0)
        unscaled_epsilon = Numeric(value_epsilon)
        utility = self.utility_expression_one_alternative(
            the_id=alternative_id,
            the_consumption=consumption,
            unscaled_epsilon=unscaled_epsilon,
        )
        result: FunctionOutput = get_value_and_derivatives(
            expression=utility,
            database=one_row,
            gradient=True,
            named_results=True,
            numerically_safe=False,
        )

        # Validate the utility calculation

        calculated_utility_value = self.utility_one_alternative(
            the_id=alternative_id,
            the_consumption=value_consumption,
            epsilon=value_epsilon,
            one_observation=one_row,
        )

        if not np.isclose(result.function, calculated_utility_value):
            error_msg = f'Inconsistent utility values for alt. {alternative_id}: {result.function} and {calculated_utility_value}'
            error_messages.append(error_msg)

        # Validate the derivative

        calculated_derivative = self.derivative_utility_one_alternative(
            the_id=alternative_id,
            the_consumption=value_consumption,
            epsilon=value_epsilon,
            one_observation=one_row,
        )

        the_gradient = result.gradient['consumption']
        if not np.isclose(the_gradient, calculated_derivative):
            error_msg = f'Inconsistent derivatives for alt. {alternative_id}: {the_gradient} and {calculated_derivative}'
            error_messages.append(error_msg)

        return error_messages

    def validation_one_alternative(
        self, alternative_id: int, one_row: Database
    ) -> list[str]:
        values = [1, 10, 100]
        error_messages = sum(
            (
                self.validation_one_alternative_one_consumption(
                    alternative_id=alternative_id,
                    value_consumption=consumption_value,
                    one_row=one_row,
                )
                for consumption_value in values
            ),
            [],
        )
        # Validate the optimal consumption

        dual_variable = 10
        value_epsilon = 0.01

        optimal_consumption = self.optimal_consumption_one_alternative(
            the_id=alternative_id,
            dual_variable=dual_variable,
            epsilon=value_epsilon,
            one_observation=one_row,
        )

        # If we calculate the derivative for this level of consumption, we should find the dual variable.
        calculated_derivative = self.derivative_utility_one_alternative(
            the_id=alternative_id,
            the_consumption=optimal_consumption,
            epsilon=value_epsilon,
            one_observation=one_row,
        )
        if not np.isclose(calculated_derivative, dual_variable):
            error_msg = (
                f'Inconsistent dual variables for alt. {alternative_id}: '
                f'{calculated_derivative} and {dual_variable}'
            )
            error_messages.append(error_msg)

        return error_messages

    def validation(self, one_row: Database) -> list[str]:
        """Validation of some abstract methods for all alternatives."""
        return sum(
            (
                self.validation_one_alternative(alternative_id=the_id, one_row=one_row)
                for the_id in self.alternatives
            ),
            [],
        )

    def forecast(
        self,
        database: Database,
        total_budget: float,
        epsilons: list[np.ndarray],
        brute_force: bool = False,
        tolerance_dual: float = 1.0e-10,
        tolerance_budget: float = 1.0e-10,
    ) -> list[pd.DataFrame]:
        """Forecast the optimal expenditures given a budget and one realization of epsilon for each alternative

        :param database: database containing the values of the explanatory variables.
        :param total_budget: total budget.
        :param epsilons: draws from the error terms provided by the user. Each entry of the list corresponds
             to an observation in the database, and consists of a data frame of
             size=(number_of_draws, self.number_of_alternatives).
        :param brute_force: if True, the brute force algorithm is applied. It solves each optimization problem using
            the scipy optimization algorithm. If False, the method proposed by Pinjari and Bhat (2021) is used.
        :param tolerance_dual: convergence criterion for the estimate of the dual variable.
        :param tolerance_budget: convergence criterion for the estimate of the total budget.
        :return: a list of data frames, each containing the results for each draw

        .. [PinjBhat21] A. R. Pinjari, C. Bhat, Computationally efficient forecasting procedures for Kuhn-Tucker
            consumer demand model systems: Application to residential energy consumption analysis, Journal of Choice
            Modelling, Volume 39, 2021, 100283.
        """

        rows_of_database = [
            Database(name=f'row_{i}', dataframe=database.data.iloc[[i]])
            for i in range(len(database.data))
        ]
        if len(rows_of_database) == 0:
            error_msg = 'Empty database'
            raise BiogemeError(error_msg)

        if len(epsilons) != len(rows_of_database):
            error_msg = (
                f'User provided draws have {len(epsilons)} entries while there are '
                f'{len(rows_of_database)} observations in the sample.'
            )
            raise BiogemeError(error_msg)

        number_of_draws = None
        for index, epsilon in enumerate(epsilons):
            if number_of_draws is None:
                number_of_draws = epsilon.shape[0]
            elif epsilon.shape[0] != number_of_draws:
                error_msg = (
                    f'User defined draws for obs. {index} contains {epsilon.shape[0]} '
                    f'draws instead of {number_of_draws}'
                )
                raise error_msg
            if epsilon.shape[1] != self.number_of_alternatives:
                error_msg = (
                    f'User defined draws for obs. {index} contains {epsilon.shape[1]} '
                    f'alternatives instead of {self.number_of_alternatives}'
                )
                raise error_msg

        all_results = []
        for index, row in enumerate(rows_of_database):
            logger.info(
                f'Forecasting observation {index} / {len(rows_of_database)} [{number_of_draws} draws]'
            )

            # Forecasting

            optimal_expenditures = (
                [
                    self.forecast_bruteforce_one_draw(
                        one_row_database=row, total_budget=total_budget, epsilon=epsilon
                    )
                    for epsilon in epsilons[index]
                ]
                if brute_force
                else [
                    self.forecast_bisection_one_draw(
                        one_row_of_database=row,
                        total_budget=total_budget,
                        epsilon=epsilon,
                        tolerance_budget=tolerance_budget,
                        tolerance_dual=tolerance_dual,
                    )
                    for epsilon in epsilons[index]
                ]
            )

            gather_data = defaultdict(list)

            # Collecting values for each key
            for optimal_solution in optimal_expenditures:
                if optimal_solution is not None:
                    for key, value in optimal_solution.items():
                        gather_data[key].append(value)

            all_results.append(pd.DataFrame(gather_data).sort_index(axis='columns'))
        return all_results

    def forecast_comparison_one_draw(
        self,
        one_row_of_database: Database,
        total_budget: float,
        epsilon: np.array,
        tolerance_dual: float = 1.0e-10,
        tolerance_budget: float = 1.0e-10,
    ) -> dict[int, float] | None:
        """Compares the results from each of the two forecasting algorithms.
        :param one_row_of_database: list of rows from the database containing the values of the explanatory variables.
        :param total_budget: total budget.
        :param epsilon: draws from the error terms provided by the user. Length: self.number_of_alternatives.
        :param tolerance_dual: convergence criterion for the estimate of the dual variable.
        :param tolerance_budget: convergence criterion for the estimate of the total budget.
        :return: None. Warning are triggered if discrepancies are observed.
        """
        if len(epsilon) != self.number_of_alternatives:
            error_msg = f'There are {self.number_of_alternatives} alternatives and {len(epsilon)} epsilons'
            raise BiogemeError(error_msg)

        logger.info('============ Comparison ===================')
        brute_force: dict[int, float] | None = self.forecast_bruteforce_one_draw(
            one_row_database=one_row_of_database,
            total_budget=total_budget,
            epsilon=epsilon,
        )

        consumption_brute_force = (
            np.array([value for key, value in sorted(brute_force.items())])
            if brute_force is not None
            else None
        )
        obj_brute_force = (
            self.sum_of_utilities(
                consumptions=consumption_brute_force,
                epsilon=epsilon,
                data_row=one_row_of_database,
            )
            if brute_force is not None
            else None
        )
        constraint_brute_force = (
            sum(consumption_brute_force) if brute_force is not None else None
        )
        choice_set_brute_force: set[int] | None = None
        if brute_force is None:
            logger.info('Brute force algorithm failed')
        else:
            choice_set_brute_force = set(
                [key for key, value in brute_force.items() if not np.isclose(value, 0)]
            )
            formatted_brute_force = {
                k: f'{value:.3g}' for k, value in brute_force.items()
            }
            logger.info(
                f'Brute force: {formatted_brute_force} objective {obj_brute_force:.3g}, constraint {constraint_brute_force:.3g}, choice set {choice_set_brute_force}'
            )
        analytical = self.forecast_bisection_one_draw(
            one_row_of_database=one_row_of_database,
            total_budget=total_budget,
            epsilon=epsilon,
            tolerance_budget=tolerance_budget,
            tolerance_dual=tolerance_dual,
        )
        choice_set_analytical: set[int] = set(
            [key for key, value in analytical.items() if not np.isclose(value, 0)]
        )
        consumption_analytical = (
            np.array([value for key, value in sorted(analytical.items())])
            if analytical is not None
            else None
        )
        try:
            obj_analytical = (
                self.sum_of_utilities(
                    consumptions=consumption_analytical,
                    epsilon=epsilon,
                    data_row=one_row_of_database,
                )
                if analytical is not None
                else None
            )
        except RuntimeWarning as e:
            logger.warning(f'{analytical=}')
            logger.warning(f'{consumption_analytical=}')
            raise e

        constraint_analytical = (
            sum(consumption_analytical) if analytical is not None else None
        )
        if analytical is None:
            logger.info('Analytical algorithm failed')
        else:
            formatted_analytical = {
                k: f'{value:.3g}' for k, value in analytical.items()
            }
            logger.info(
                f'Analytical:  {formatted_analytical} objective {obj_analytical:.3g}, constraint {constraint_analytical:.3g}, choice set {choice_set_analytical}'
            )
        if brute_force is None and analytical is None:
            logger.warning('Both algorithms failed.')
            return
        if brute_force is None:
            logger.warning('Brute force algorithm failed.')
            return
        if analytical is None:
            logger.warning('Analytical algorithm failed.')
            return
        if (
            choice_set_brute_force != choice_set_analytical
            and choice_set_brute_force is not None
        ):
            logger.warning(
                f'Different optimal choice sets: analytical {choice_set_analytical}, brute force {choice_set_brute_force}'
            )
        if not np.isclose(obj_analytical, obj_brute_force):
            logger.warning(
                f'Difference between optimal utility with analytical [{obj_analytical:.4g}] and brute '
                f'force [{obj_brute_force:.4g}] algorithms.'
            )
            brute_force_results = ', '.join(
                [f'{key}: {value:.3g}' for key, value in brute_force.items()]
            )
            analytical_results = ', '.join(
                [f'{key}: {value:.3g}' for key, value in analytical.items()]
            )
            logger.warning(f'Solution with brute force: {brute_force_results}')
            logger.warning(f'Solution with analytical: {analytical_results}')
        if not np.isclose(constraint_analytical, constraint_brute_force):
            logger.warning(
                f'Difference between constraint with analytical [{constraint_analytical}] and brute force '
                f'[{constraint_brute_force}] algorithms.'
            )

    def generate_epsilons(
        self, number_of_observations: int, number_of_draws: int
    ) -> list[np.ndarray]:
        """Generate draws for the error terms.

        :param number_of_observations: number of entries in the database.
        :param number_of_draws: number of draws to generate
        :return: a list of length "number_of_observations". Each element is a
        number_of_draws x self.number_of_alternatives array.
        """
        return [
            np.random.gumbel(
                loc=0, scale=1, size=(number_of_draws, self.number_of_alternatives)
            )
            for _ in range(number_of_observations)
        ]

    def validate_forecast(
        self,
        database: Database,
        total_budget: float,
        epsilons: list[np.ndarray],
        tolerance_dual: float = 1.0e-13,
        tolerance_budget: float = 1.0e-13,
    ) -> None:
        """Compare the two algorithms to forecast the optimal expenditures given a budget and one realization of
        epsilon for each alternative

        :param database: database containing the values of the explanatory variables.
        :param total_budget: total budget.
        :param epsilons: draws from the error terms. Each entry of the list corresponds
             to an observation in the database, and consists of a data frame of
             size=(number_of_draws, self.number_of_alternatives).
        :param tolerance_dual: convergence criterion for the estimate of the dual variable.
        :param tolerance_budget: convergence criterion for the estimate of the total budget.

        :return: None. Warning are triggered if discrepancies are observed

        """

        rows_of_database = database.mdcev_row_split()
        if len(rows_of_database) == 0:
            error_msg = 'Empty database'
            raise BiogemeError(error_msg)

        if len(epsilons) != len(rows_of_database):
            error_msg = (
                f'User provided draws have {len(epsilons)} entries while there are '
                f'{len(rows_of_database)} observations in the sample.'
            )
            raise BiogemeError(error_msg)

        number_of_draws = len(epsilons[0])
        for index, row in enumerate(rows_of_database):
            logger.info(
                f'Forecasting observation {index} / {len(rows_of_database)} [{number_of_draws} draws]'
            )

            # Forecasting
            for epsilon in epsilons[index]:
                self.forecast_comparison_one_draw(
                    total_budget=total_budget,
                    one_row_of_database=row,
                    epsilon=epsilon,
                    tolerance_budget=tolerance_budget,
                    tolerance_dual=tolerance_dual,
                )



"""

Mixture with lognormal distribution
===================================

Example of a mixture of logit models.  The mixing distribution is
distributed as a log normal.  Compared to
:ref:`plot_b17lognormal_mixture`, the integration is performed using
numerical integration instead of Monte-Carlo approximation.

:author: Michel Bierlaire, EPFL
:date: Mon Apr 10 12:13:23 2023

"""

import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
import biogeme.distributions as dist
from biogeme import models
from biogeme.expressions import (
    Beta,
    RandomVariable,
    exp,
    log,
    Integrate,
)

# %%
# See the data processing script: :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b17lognormal_mixture_integral.py')

# %%
# Parameters to be estimated.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_COST = Beta('B_COST', 0, None, None, 0)

# %%
# Define a random parameter, normally distributed, designed to be used.
# for Monte-Carlo simulation
B_TIME = Beta('B_TIME', 0, None, None, 0)

# %%
# It is advised not to use 0 as starting value for the following parameter..
B_TIME_S = Beta('B_TIME_S', 1, -2, 2, 0)

# %%
# Define a random parameter, log normally distributed, designed to be used
# for numerical integration.
omega = RandomVariable('omega')
B_TIME_RND = -exp(B_TIME + B_TIME_S * omega)
density = dist.normalpdf(omega)

# %%
# Definition of the utility functions.
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Conditional to omega, we have a logit model (called the kernel).
condprob = models.logit(V, av, CHOICE)

# %%
# We integrate over omega using numerical integration.
logprob = log(Integrate(condprob * density, 'omega'))

# %%
# Create the Biogeme object.
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b17lognormal_mixture_integral'

# %%
# Estimate the parameters
results = the_biogeme.estimate()

# %%
print(results.short_summary())

# %%
pandas_results = results.getEstimatedParameters()
pandas_results



""".. _plot_b22multiple_models:

Assisted specification
======================

Example of the estimation of several versions of the model using
assisted specification algorithm. The catalog of specifications is
defined in :ref:`plot_b22multiple_models_spec` . Compared to
:ref:`plot_b21multiple_models`, the number fo specifications exceeds
the maximum limit, so a heuristic is applied.  See `Bierlaire and
Ortelli, 2023
<https://transp-or.epfl.ch/documents/technicalReports/BierOrte23.pdf>`_
for a detailed description of the use of the assisted specification
algorithm.

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 17:05:40 2023


"""

import biogeme.biogeme_logging as blog
from biogeme.results import compile_estimation_results
from biogeme.multiobjectives import AIC_BIC_dimension
from biogeme.assisted import AssistedSpecification
from plot_b22multiple_models_spec import the_biogeme, PARETO_FILE_NAME


logger = blog.get_screen_logger(blog.INFO)
logger.info('Example b22multiple_models')

# %%
nbr = the_biogeme.loglike.number_of_multiple_expressions()
if nbr is None:
    print('There are too many possible specifications to be enumerated')
else:
    print(f'There are {nbr} possible specifications')


# %%
# Creation of the object capturing the assisted specification algorithm.
# Its constructor takes three arguments:
#
#    - the biogeme object containing the specifications and the
#      database,
#    - an object defining the objectives to minimize. Here, we use
#      three objectives: AIC, BIC and number of parameters.
#    - the name of the file where the estimated are saved, and
#      organized into a Pareto set.
assisted_specification = AssistedSpecification(
    biogeme_object=the_biogeme,
    multi_objectives=AIC_BIC_dimension,
    pareto_file_name=PARETO_FILE_NAME,
)

# %%
# The algorithm is run.
non_dominated_models = assisted_specification.run()

# %%
summary, description = compile_estimation_results(
    non_dominated_models, use_short_names=True
)
print(summary)

# %%
# Explanation of the short names of the model.
for k, v in description.items():
    if k != v:
        print(f'{k}: {v}')



"""

Choice model with the latent variable: maximum likelihood estimation
====================================================================

Mixture of logit.
Measurement equation for the indicators.
Maximum likelihood (full information) estimation.

:author: Michel Bierlaire, EPFL
:date: Fri Apr 14 10:07:43 2023
"""

import sys
from functools import reduce
import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
import biogeme.exceptions as excep
import biogeme.distributions as dist
import biogeme.results as res
from biogeme import models
from biogeme.expressions import (
    Beta,
    Variable,
    log,
    RandomVariable,
    Integrate,
    Elem,
    bioNormalCdf,
    exp,
)

from read_or_estimate import read_or_estimate

from optima import (
    database,
    male,
    age,
    haveChildren,
    highEducation,
    childCenter,
    childSuburb,
    SocioProfCat,
    WaitingTimePT,
    Choice,
    TimePT_scaled,
    TimeCar_scaled,
    MarginalCostPT_scaled,
    CostCarCHF_scaled,
    distance_km_scaled,
    PurpHWH,
    PurpOther,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example m03_simultaneous_estimation.py')

# %%
# Read the estimates from the structural equation estimation.
MODELNAME = 'm01_latent_variable'
try:
    struct_results = res.bioResults(pickleFile=f'saved_results/{MODELNAME}.pickle')
except excep.BiogemeError:
    print(
        f'Run first the script {MODELNAME}.py in order to generate the '
        f'file {MODELNAME}.pickle, and move it to the directory saved_results'
    )
    sys.exit()
struct_betas = struct_results.getBetaValues()


# %%
# Coefficients
coef_intercept = Beta('coef_intercept', struct_betas['coef_intercept'], None, None, 0)
coef_age_30_less = Beta(
    'coef_age_30_less', struct_betas['coef_age_30_less'], None, None, 0
)
coef_haveChildren = Beta(
    'coef_haveChildren', struct_betas['coef_haveChildren'], None, None, 0
)
coef_highEducation = Beta(
    'coef_highEducation', struct_betas['coef_highEducation'], None, None, 0
)
coef_artisans = Beta('coef_artisans', struct_betas['coef_artisans'], None, None, 0)
coef_employees = Beta('coef_employees', struct_betas['coef_employees'], None, None, 0)
coef_male = Beta('coef_male', struct_betas['coef_male'], None, None, 0)
coef_child_center = Beta(
    'coef_child_center', struct_betas['coef_child_center'], None, None, 0
)
coef_child_suburb = Beta(
    'coef_child_suburb', struct_betas['coef_child_suburb'], None, None, 0
)

# %%
# Latent variable: structural equation

# %%
# Define a random parameter, normally distributed, designed to be used
# for numerical integration
omega = RandomVariable('omega')
density = dist.normalpdf(omega)
sigma_s = Beta('sigma_s', 1, None, None, 0)

# %%
ACTIVELIFE = (
    coef_intercept
    + coef_child_center * childCenter
    + coef_child_suburb * childSuburb
    + coef_highEducation * highEducation
    + coef_artisans * (SocioProfCat == 5)
    + coef_employees * (SocioProfCat == 6)
    + coef_age_30_less * (age <= 30)
    + coef_male * male
    + coef_haveChildren * haveChildren
    + sigma_s * omega
)

# %%
# Measurement equations
indicators = [
    'ResidCh01',
    'ResidCh04',
    'ResidCh05',
    'ResidCh06',
    'LifSty07',
    'LifSty10',
]

# %%
# We define the intercept parameters. The first one is normalized to 0.
inter = {k: Beta(f'inter_{k}', 0, None, None, 0) for k in indicators[1:]}
inter[indicators[0]] = Beta(f'INTER_{indicators[0]}', 0, None, None, 1)

# %%
# We define the coefficients. The first one is normalized to 1.
coefficients = {k: Beta(f'coeff_{k}', 0, None, None, 0) for k in indicators[1:]}
coefficients[indicators[0]] = Beta(f'B_{indicators[0]}', 1, None, None, 1)

# %%
# We define the measurement equations for each indicator
linear_models = {k: inter[k] + coefficients[k] * ACTIVELIFE for k in indicators}

# %%
# We define the scale parameters of the error terms.
sigma_star = {k: Beta(f'sigma_star_{k}', 1, 1.0e-5, None, 0) for k in indicators[1:]}
sigma_star[indicators[0]] = Beta(f'sigma_star_{indicators[0]}', 1, None, None, 1)

# %%
# Symmetric threshold.
delta_1 = Beta('delta_1', 0.1, 1.0e-5, None, 0)
delta_2 = Beta('delta_2', 0.2, 1.0e-5, None, 0)
tau_1 = -delta_1 - delta_2
tau_2 = -delta_1
tau_3 = delta_1
tau_4 = delta_1 + delta_2

# %%
# Ordered probit models.
tau_1_residual = {k: (tau_1 - linear_models[k]) / sigma_star[k] for k in indicators}
tau_2_residual = {k: (tau_2 - linear_models[k]) / sigma_star[k] for k in indicators}
tau_3_residual = {k: (tau_3 - linear_models[k]) / sigma_star[k] for k in indicators}
tau_4_residual = {k: (tau_4 - linear_models[k]) / sigma_star[k] for k in indicators}
dict_prob_indicators = {
    k: {
        1: bioNormalCdf(tau_1_residual[k]),
        2: bioNormalCdf(tau_2_residual[k]) - bioNormalCdf(tau_1_residual[k]),
        3: bioNormalCdf(tau_3_residual[k]) - bioNormalCdf(tau_2_residual[k]),
        4: bioNormalCdf(tau_4_residual[k]) - bioNormalCdf(tau_3_residual[k]),
        5: 1 - bioNormalCdf(tau_4_residual[k]),
        6: 1.0,
        -1: 1.0,
        -2: 1.0,
    }
    for k in indicators
}

# %%
# Product of the likelihood of the indicators.
prob_indicators = reduce(
    lambda x, y: x * Elem(dict_prob_indicators[y], Variable(y)),
    indicators,
    Elem(dict_prob_indicators[indicators[0]], Variable(indicators[0])),
)


# %%
# Choice model
# Read the estimates from the sequential estimation, and use
# them as starting values
MODELNAME = 'm02_sequential_estimation'
try:
    choice_results = res.bioResults(pickleFile=f'saved_results/{MODELNAME}.pickle')
except excep.BiogemeError:
    print(
        f'Run first the script {MODELNAME}.py in order to generate the '
        f'file {MODELNAME}.pickle, and move it to the directory saved_results'
    )
    sys.exit()
choice_betas = choice_results.getBetaValues()

# %%
ASC_CAR = Beta('ASC_CAR', choice_betas['ASC_CAR'], None, None, 0)
ASC_PT = Beta('ASC_PT', 0, None, None, 1)
ASC_SM = Beta('ASC_SM', choice_betas['ASC_SM'], None, None, 0)
BETA_COST_HWH = Beta('BETA_COST_HWH', choice_betas['BETA_COST_HWH'], None, None, 0)
BETA_COST_OTHER = Beta(
    'BETA_COST_OTHER', choice_betas['BETA_COST_OTHER'], None, None, 0
)
BETA_DIST = Beta('BETA_DIST', choice_betas['BETA_DIST'], None, None, 0)
BETA_TIME_CAR_REF = Beta(
    'BETA_TIME_CAR_REF', choice_betas['BETA_TIME_CAR_REF'], None, 0, 0
)
BETA_TIME_CAR_AL = Beta(
    'BETA_TIME_CAR_AL', choice_betas['BETA_TIME_CAR_AL'], None, None, 0
)
BETA_TIME_PT_REF = Beta(
    'BETA_TIME_PT_REF', choice_betas['BETA_TIME_PT_REF'], None, 0, 0
)
BETA_TIME_CAR = BETA_TIME_CAR_REF * exp(BETA_TIME_CAR_AL * ACTIVELIFE)
BETA_TIME_PT_AL = Beta(
    'BETA_TIME_PT_AL', choice_betas['BETA_TIME_PT_AL'], None, None, 0
)
BETA_TIME_PT = BETA_TIME_PT_REF * exp(BETA_TIME_PT_AL * ACTIVELIFE)
BETA_WAITING_TIME = Beta(
    'BETA_WAITING_TIME', choice_betas['BETA_WAITING_TIME'], None, None, 0
)

# %%
# Definition of utility functions:
V0 = (
    ASC_PT
    + BETA_TIME_PT * TimePT_scaled
    + BETA_WAITING_TIME * WaitingTimePT
    + BETA_COST_HWH * MarginalCostPT_scaled * PurpHWH
    + BETA_COST_OTHER * MarginalCostPT_scaled * PurpOther
)

V1 = (
    ASC_CAR
    + BETA_TIME_CAR * TimeCar_scaled
    + BETA_COST_HWH * CostCarCHF_scaled * PurpHWH
    + BETA_COST_OTHER * CostCarCHF_scaled * PurpOther
)

V2 = ASC_SM + BETA_DIST * distance_km_scaled

# %%
# Associate utility functions with the numbering of alternatives
V = {0: V0, 1: V1, 2: V2}

# %%
# Conditional on omega, we have a logit model (called the kernel) for
# the choice
condprob = models.logit(V, None, Choice)

# %%
# Conditional on omega, we have the product of ordered probit for the
# indicators.
condlike = prob_indicators * condprob

# %%
# We integrate over omega using numerical integration
loglike = log(Integrate(condlike * density, 'omega'))

# %%
# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, loglike)
the_biogeme.modelName = 'm03_simultaneous_estimation'

# %%
# If estimation results are saved on file, we read them to speed up the process.
# If not, we estimate the parameters.
results = read_or_estimate(the_biogeme=the_biogeme, directory='saved_results')

# %%
print(results.short_summary())

# %%
print(f'Final log likelihood: {results.data.logLike:.3f}')
print(f'Output file: {results.data.htmlFileName}')

# %%
results.getEstimatedParameters()



"""Implements the function providing names for the output .py.

:author: Michel Bierlaire

:date: Tue Mar 26 16:48:40 2019
"""

import logging
from pathlib import Path

logger = logging.getLogger(__name__)


def get_new_file_name(name: str, ext: str) -> str:
    """
      Generate a file name that does not exist.

    :param name: name of the file.

    :param ext: file extension.

    :return: name.ext if the file does not exist.  If it does, returns
       name~xx.ext, where xx is the smallest integer such that the
       corresponding file does not exist. It is designed to avoid erasing
       output .py inadvertently.

    """
    file_name = name + '.' + ext
    the_file = Path(file_name)
    number = int(0)
    while the_file.is_file():
        file_name = f'{name}~{number:02d}.{ext}'
        the_file = Path(file_name)
        number += 1
    return file_name



"""
MDCEV Utilities

This module provides helper functions for preparing and transforming
data specifically for MDCEV (Multiple Discrete Continuous Extreme Value) models.

Michel Bierlaire
Wed Mar 26 19:43:21 2025
"""

from typing import Iterable
import pandas as pd
from biogeme.database import Database


def mdcev_count(
    df: pd.DataFrame, list_of_columns: list[str], new_column: str
) -> pd.DataFrame:
    """
    Computes the number of non-zero entries across specified columns,
    corresponding to the number of goods consumed in MDCEV.

    :param df: DataFrame containing the MDCEV data.
    :param list_of_columns: Columns representing quantities of each good.
    :param new_column: Name of the output column to store the count.
    :return: Modified DataFrame with the count column added.
    """
    df[new_column] = df[list_of_columns].apply(lambda row: (row != 0).sum(), axis=1)
    return df


def mdcev_row_split(
    df: pd.DataFrame, a_range: Iterable[int] | None = None
) -> list[Database]:
    """
    Splits a DataFrame into a list of Database objects, one for each row,
    useful for row-level MDCEV processing.

    :param df: DataFrame to split.
    :param a_range: Optional subset of row indices to extract.
    :return: List of Database objects.
    """
    if a_range is None:
        a_range = range(len(df))
    else:
        max_index = len(df) - 1
        if any(i < 0 or i > max_index for i in a_range):
            raise IndexError("One or more indices in a_range are out of bounds.")

    return [Database(name=f'row_{i}', pandas_database=df.iloc[[i]]) for i in a_range]



"""Decorators for deprecated functions and parameters

Michel Bierlaire
Mon Jun 17 20:50:19 2024
"""

import inspect
import os
import warnings
import functools
from typing import Callable, Any

from biogeme.exceptions import BiogemeError

RAISE_EXCEPTION = False

if os.environ.get('TOX_ENV_NAME') is not None and RAISE_EXCEPTION:
    raise Exception(
        "Exception raised during testing with tox. Remove the exception raised by the deprecated functions"
    )


def deprecated(
    new_func: Callable[..., Any]
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Decorator for deprecated functions. It calls the new version of the function

    :param new_func:
    :return:
    """

    def decorator(old_func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(old_func)
        def wrapper(*args, **kwargs) -> Any:
            msg = f"{old_func.__name__} is deprecated; use {new_func.__name__} instead."
            if RAISE_EXCEPTION:
                raise BiogemeError('Deprecated')
            warnings.warn(msg, DeprecationWarning, stacklevel=2)
            return new_func(*args, **kwargs)  # Directly call the new function

        wrapper.__deprecated__ = True
        wrapper.__newname__ = new_func.__name__
        return wrapper

    return decorator


def deprecated_parameters(obsolete_params: dict[str, str | None]):

    def decorator(func):
        func_signature = inspect.signature(func)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            processed_kwargs = {}
            for name, value in list(kwargs.items()):
                if name in obsolete_params:
                    new_name = obsolete_params[name]
                    if new_name:
                        warnings.warn(
                            f"Parameter '{name}' is deprecated; use '{new_name}={value}' instead.",
                            DeprecationWarning,
                            stacklevel=2,
                        )
                        # raise BiogemeError(f'Deprecated {name}. Use {new_name}')
                        processed_kwargs[new_name] = value
                    else:
                        warnings.warn(
                            f"Parameter '{name}' is deprecated and is ignored. "
                            f"It will be removed in a future version.",
                            DeprecationWarning,
                            stacklevel=2,
                        )
                else:
                    processed_kwargs[name] = value

            return func(*args, **processed_kwargs)

        return wrapper

    return decorator


def issue_deprecation_warning(text: str) -> None:
    warnings.warn(
        text,
        DeprecationWarning,
        stacklevel=2,
    )



"""Tests for the controller class

Michel Bierlaire
Thu Apr 17 2025, 08:34:51
"""

import unittest
from biogeme.catalog import Controller
from biogeme.exceptions import BiogemeError
from biogeme.expressions import SELECTION_SEPARATOR


class DummyCatalog:
    """A minimal dummy catalog with a current_index attribute."""

    def __init__(self):
        self.current_index = -1


class TestController(unittest.TestCase):

    def setUp(self):
        self.names = ['spec1', 'spec2', 'spec3']
        self.controller = Controller('myCtrl', self.names)

    def test_constructor_stores_values(self):
        self.assertEqual(self.controller.controller_name, 'myCtrl')
        self.assertEqual(self.controller.specification_names, tuple(self.names))
        self.assertEqual(self.controller.current_index, 0)
        self.assertEqual(self.controller.dict_of_index['spec2'], 1)

    def test_all_configurations(self):
        configs = self.controller.all_configurations()
        expected = {f'myCtrl{SELECTION_SEPARATOR}{name}' for name in self.names}
        self.assertEqual(configs, expected)

    def test_current_name(self):
        self.controller.set_index(2)
        self.assertEqual(self.controller.current_name(), 'spec3')

    def test_set_name_valid(self):
        self.controller.set_name('spec2')
        self.assertEqual(self.controller.current_index, 1)

    def test_set_name_invalid_raises(self):
        with self.assertRaisesRegex(BiogemeError, 'unknown specification'):
            self.controller.set_name('nonexistent')

    def test_set_index_valid_and_updates_catalogs(self):
        dummy = DummyCatalog()
        self.controller.controlled_catalogs.append(dummy)
        self.controller.set_index(1)
        self.assertEqual(self.controller.current_index, 1)
        self.assertEqual(dummy.current_index, 1)

    def test_set_index_out_of_range_low(self):
        with self.assertRaisesRegex(BiogemeError, 'Wrong index -1'):
            self.controller.set_index(-1)

    def test_set_index_out_of_range_high(self):
        with self.assertRaisesRegex(BiogemeError, 'Wrong index 3'):
            self.controller.set_index(3)

    def test_reset_selection(self):
        self.controller.set_index(2)
        self.controller.reset_selection()
        self.assertEqual(self.controller.current_index, 0)

    def test_modify_controller_step_forward_circular(self):
        self.controller.set_index(2)
        self.controller.modify_controller(1, circular=True)
        self.assertEqual(self.controller.current_index, 0)

    def test_modify_controller_step_backward_circular(self):
        self.controller.set_index(0)
        self.controller.modify_controller(-1, circular=True)
        self.assertEqual(self.controller.current_index, 2)

    def test_modify_controller_step_forward_non_circular(self):
        self.controller.set_index(2)
        mod = self.controller.modify_controller(1, circular=False)
        self.assertEqual(self.controller.current_index, 2)
        self.assertEqual(mod, 0)

    def test_modify_controller_step_backward_non_circular(self):
        self.controller.set_index(1)
        mod = self.controller.modify_controller(-2, circular=False)
        self.assertEqual(self.controller.current_index, 0)
        self.assertEqual(mod, 1)

    def test_eq_and_hash(self):
        c1 = Controller('A', ['x'])
        c2 = Controller('A', ['y'])
        c3 = Controller('B', ['x'])
        self.assertEqual(c1, c2)
        self.assertNotEqual(c1, c3)
        self.assertEqual(hash(c1), hash(c2))
        self.assertNotEqual(hash(c1), hash(c3))

    def test_repr_and_str(self):
        r = repr(self.controller)
        s = str(self.controller)
        self.assertEqual(r, repr('myCtrl'))
        self.assertIn('myCtrl', s)
        self.assertIn('spec1', s)


if __name__ == '__main__':
    unittest.main()



"""

biogeme.optimization
====================

Examples of use of several functions.

This is designed for programmers who need examples of use of the
functions of the module. The examples are designed to illustrate the
syntax. They do not correspond to any meaningful model.

:author: Michel Bierlaire
:date: Thu Nov 23 22:25:13 2023
"""


import pandas as pd
from biogeme.version import getText
import biogeme.biogeme as bio
import biogeme.database as db
from biogeme import models
from biogeme.expressions import Beta, Variable
import biogeme.biogeme_logging as blog


# %%
# Version of Biogeme.
print(getText())

# %%
logger = blog.get_screen_logger(blog.INFO)

# %%
# Data,
df = pd.DataFrame(
    {
        'Person': [1, 1, 1, 2, 2],
        'Exclude': [0, 0, 1, 0, 1],
        'Variable1': [1, 2, 3, 4, 5],
        'Variable2': [10, 20, 30, 40, 50],
        'Choice': [1, 2, 3, 1, 2],
        'Av1': [0, 1, 1, 1, 1],
        'Av2': [1, 1, 1, 1, 1],
        'Av3': [0, 1, 1, 1, 1],
    }
)
df

# %%
my_data = db.Database('test', df)

# %%
# Variables.
Choice = Variable('Choice')
Variable1 = Variable('Variable1')
Variable2 = Variable('Variable2')
beta1 = Beta('beta1', 0, None, None, 0)
beta2 = Beta('beta2', 0, None, None, 0)
V1 = beta1 * Variable1
V2 = beta2 * Variable2
V3 = 0
V = {1: V1, 2: V2, 3: V3}

# %%
likelihood = models.loglogit(V, av=None, i=Choice)
my_biogeme = bio.BIOGEME(my_data, likelihood)
my_biogeme.modelName = 'simpleExample'
my_biogeme.save_iterations = False
my_biogeme.generate_html = False
my_biogeme.generate_pickle = False
print(my_biogeme)

# %%
f, g, h, gdiff, hdiff = my_biogeme.checkDerivatives(
    beta=my_biogeme.beta_values_dict_to_list(), verbose=True
)

# %%
pd.DataFrame(gdiff)

# %%
pd.DataFrame(hdiff)

# %%
# **scipy**: this is the optimization algorithm from scipy.
my_biogeme.algorithm_name = 'scipy'
results = my_biogeme.estimate()
results.getEstimatedParameters()

# %%
for k, v in results.data.optimizationMessages.items():
    print(f'{k}:\t{v}')

# %%
# **Newton with linesearch**
my_biogeme.algorithm_name = 'LS-newton'
results = my_biogeme.estimate()
results.getEstimatedParameters()

# %%
for k, v in results.data.optimizationMessages.items():
    print(f'{k}:\t{v}')

# %%
# Changing the requested precision
my_biogeme.tolerance = 0.1
results = my_biogeme.estimate()
results.getEstimatedParameters()

# %%
for k, v in results.data.optimizationMessages.items():
    print(f'{k}:\t{v}')

# %%
# **Newton with trust region**
my_biogeme.algorithm_name = 'TR-newton'
results = my_biogeme.estimate()
results.getEstimatedParameters()

# %%
for k, v in results.data.optimizationMessages.items():
    print(f'{k}:\t{v}')

# %%
# We illustrate the parameters. We use the truncated conjugate
# gradient instead of dogleg for the trust region subproblem, starting
# with a small trust region of radius 0.001, and a maximum of 3
# iterations.

# %%
my_biogeme.dogleg = False
my_biogeme.initial_radius = 0.001
my_biogeme.maxiter = 3
results = my_biogeme.estimate()
results.getEstimatedParameters()

# %%
for k, v in results.data.optimizationMessages.items():
    print(f'{k}:\t{v}')

# %%
# Changing the requested precision
my_biogeme.tolerance = 0.1
my_biogeme.maxiter = 1000
results = my_biogeme.estimate()
results.getEstimatedParameters()

# %%
for k, v in results.data.optimizationMessages.items():
    print(f'{k}:\t{v}')

# %%
# **BFGS with line search**
my_biogeme.algorithm_name = 'LS-BFGS'
my_biogeme.tolerance = 1.0e-6
my_biogeme.maxiter = 1000
results = my_biogeme.estimate()
results.getEstimatedParameters()

# %%
for k, v in results.data.optimizationMessages.items():
    print(f'{k}:\t{v}')

# %%
# **BFGS with trust region**
my_biogeme.algorithm_name = 'TR-BFGS'
results = my_biogeme.estimate()
results.getEstimatedParameters()

# %%
for k, v in results.data.optimizationMessages.items():
    print(f'{k}:\t{v}')

# %%
# **Newton/BFGS with trust region for simple bounds**

# %%
# This is the default algorithm used by Biogeme. It is the
# implementation of the algorithm proposed by `Conn et al. (1988)
# <https://www.ams.org/journals/mcom/1988-50-182/S0025-5718-1988-0929544-3/S0025-5718-1988-0929544-3.pdf>`_.
my_biogeme.algorithm_name = 'simple_bounds'
results = my_biogeme.estimate()
results.getEstimatedParameters()

# %%
for k, v in results.data.optimizationMessages.items():
    print(f'{k}:\t{v}')

# %%
# When the second derivatives are too computationally expensive to
# calculate, it is possible to avoid calculating them at each
# successful iteration. The parameter `second_derivatives` allows to
# control that.

# %%
my_biogeme.second_derivatives = 0.5
results = my_biogeme.estimate()
results.getEstimatedParameters()

# %%
for k, v in results.data.optimizationMessages.items():
    print(f'{k}:\t{v}')

# %%
# If the parameter is set to zero, the second derivatives are not used
# at all, and the algorithm relies only on the BFGS update.

# %%
my_biogeme.second_derivatives = 0.0
results = my_biogeme.estimate()
results.getEstimatedParameters()

# %%
for k, v in results.data.optimizationMessages.items():
    print(f'{k}:\t{v}')

# %%
# There are shortcuts to call the BFGS and the Newton versions
my_biogeme.algorithm_name = 'simple_bounds_newton'
results = my_biogeme.estimate()
results.getEstimatedParameters()

# %%
for k, v in results.data.optimizationMessages.items():
    print(f'{k}:\t{v}')

# %%
my_biogeme.algorithm_name = 'simple_bounds_BFGS'
results = my_biogeme.estimate()
results.getEstimatedParameters()

# %%
for k, v in results.data.optimizationMessages.items():
    print(f'{k}:\t{v}')



"""
Test the biogeme module

:author: Michel Bierlaire
:date: Wed Apr 29 18:32:42 2020

"""

import os
import random as rnd
import unittest

import numpy as np

import biogeme.biogeme_logging as blog
from biogeme.biogeme import BIOGEME
from biogeme.catalog import Catalog
from biogeme.exceptions import BiogemeError
from biogeme.expressions import (
    Beta,
    Numeric,
    PanelLikelihoodTrajectory,
    RandomVariable,
    Variable,
    exp,
)
from biogeme.second_derivatives import SecondDerivativesMode
from biogeme.tools.files import files_of_type
from biogeme.validation import ValidationResult
from test_data import getData, getPanelData

logger = blog.get_screen_logger(level=blog.INFO)


class TestBiogeme(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(90267)
        rnd.seed(90267)

    @classmethod
    def tearDownClass(cls):
        for file_name in os.listdir('.'):
            if (
                file_name.endswith('.pickle')
                or file_name.endswith('.html')
                or file_name.endswith('.iter')
                or file_name.endswith('.log')
                or file_name.endswith('.yaml')
            ):
                os.remove(file_name)

    def get_dict_of_expressions(self):
        Variable1 = Variable('Variable1')
        Variable2 = Variable('Variable2')
        beta1 = Beta('beta1', -1.0, -3, 3, 0)
        beta2 = Beta('beta2', 2.0, -3, 10, 0)
        likelihood = -((beta1 * Variable1) ** 2) - (beta2 * Variable2) ** 2
        simul = (beta1 + 2 * beta2) / Variable1 + (beta2 + 2 * beta1) / Variable2
        dict_of_expressions = {
            'log_like': likelihood,
            'weight': Numeric(1),
            'beta1': beta1,
            'simul': simul,
        }
        return dict_of_expressions

    def get_dict_of_expressions_for_panel(self):
        Variable1 = Variable('Variable1')
        Variable2 = Variable('Variable2')
        beta1 = Beta('beta1', -1.0, -3, 3, 0)
        beta2 = Beta('beta2', 2.0, -3, 10, 0)
        likelihood = exp(-((beta1 * Variable1) ** 2) - (beta2 * Variable2) ** 2)
        simul = (beta1 + 2 * beta2) / Variable1 + (beta2 + 2 * beta1) / Variable2
        dict_of_expressions = {
            'log_like': PanelLikelihoodTrajectory(likelihood),
            'weight': Numeric(1),
            'beta1': beta1,
            'simul': PanelLikelihoodTrajectory(simul),
        }
        return dict_of_expressions

    def get_biogeme_instance(self):
        data = getData(1)
        return BIOGEME(data, self.get_dict_of_expressions())

    def get_biogeme_instance_without_bounds(self):
        Variable1 = Variable('Variable1')
        Variable2 = Variable('Variable2')
        beta1 = Beta('beta1', -1.0, None, None, 0)
        beta2 = Beta('beta2', 2.0, None, None, 0)
        likelihood = -((beta1 * Variable1) ** 2) - (beta2 * Variable2) ** 2
        simul = (beta1 + 2 * beta2) / Variable1 + (beta2 + 2 * beta1) / Variable2
        dict_of_expressions = {
            'log_like': likelihood,
            'weight': Numeric(1),
            'beta1': beta1,
            'simul': simul,
        }
        data = getData(1)
        return BIOGEME(data, dict_of_expressions)

    def get_panel_instance(self):
        Variable1 = Variable('Variable1')
        Variable2 = Variable('Variable2')
        beta1 = Beta('beta1', -1.0, -3, 3, 0)
        beta2 = Beta('beta2', 2.0, -3, 10, 0)
        likelihood = -((beta1 * Variable1) ** 2) - (beta2 * Variable2) ** 2
        simul = (beta1 + 2 * beta2) / Variable1 + (beta2 + 2 * beta1) / Variable2
        dict_of_expressions = {
            'log_like': PanelLikelihoodTrajectory(likelihood),
            'weight': Numeric(1),
            'beta1': beta1,
            'simul': PanelLikelihoodTrajectory(simul),
        }
        data = getData(1)
        data.panel('Person')
        return BIOGEME(data, dict_of_expressions)

    def test_ctor(self):
        # Test obsolete parameters
        aBiogeme = BIOGEME(
            getData(1),
            self.get_dict_of_expressions(),
        )
        wrong_data = getData(1)
        wrong_data.dataframe.loc['Person', 0] = np.nan
        with self.assertRaises(BiogemeError):
            bBiogeme = BIOGEME(
                wrong_data,
                self.get_dict_of_expressions(),
            )

        with self.assertRaises(BiogemeError):
            bBiogeme = BIOGEME(
                getData(1),
                'wrong_object',
            )

        with self.assertRaises(BiogemeError):
            bBiogeme = BIOGEME(
                getData(1),
                {'log_like': 'wrong_object'},
            )

        wrong_expression = Variable('Variable1') * PanelLikelihoodTrajectory(
            Beta('beta1', -1.0, -3, 3, 0)
        )
        with self.assertRaises(BiogemeError):
            bBiogeme = BIOGEME(
                getPanelData(1),
                wrong_expression,
            )

        cBiogeme = BIOGEME(
            getPanelData(1),
            self.get_dict_of_expressions_for_panel(),
        )

    def test_saved_iterations(self):
        my_biogeme = self.get_biogeme_instance()
        my_biogeme._load_saved_iteration()
        # Remove the file and try to loaed it again
        my_biogeme._load_saved_iteration()

    def test_random_init_values(self):
        my_biogeme = self.get_biogeme_instance()
        my_biogeme.set_random_init_values(default_bound=10)
        for (
            v
        ) in (
            my_biogeme.model_elements.expressions_registry.list_of_free_betas_init_values
        ):
            self.assertLessEqual(v, 10)
            self.assertGreaterEqual(v, -10)

    def test_free_beta_names(self):
        my_biogeme = self.get_biogeme_instance()
        result = my_biogeme.free_betas_names
        expected_result = ['beta1', 'beta2']
        self.assertListEqual(result, expected_result)

    def test_saveIterationsFileName(self):
        my_biogeme = self.get_biogeme_instance()
        my_biogeme.model_name = 'simpleExample'
        f = my_biogeme._save_iterations_file_name()
        self.assertEqual(f, '__simpleExample.iter')

    def test_random_variable(self):
        rv = RandomVariable('omega')
        with self.assertRaises(BiogemeError):
            b = BIOGEME(getData(1), rv)

    def test_calculateNullLoglikelihood(self):
        my_biogeme = self.get_biogeme_instance()
        null_ell = my_biogeme.calculate_null_loglikelihood({1: 1, 2: 1})
        self.assertAlmostEqual(null_ell, -3.4657359027997265, 2)
        null_ell_2 = my_biogeme.calculate_null_loglikelihood({1: 1, 2: 1, 3: 1})
        self.assertAlmostEqual(null_ell_2, -5.493061443340549, 2)

    def test_calculateInitLikelihood(self):
        my_biogeme = self.get_biogeme_instance()
        res = my_biogeme.calculate_init_likelihood()
        self.assertAlmostEqual(res, -22055, 5)

    def test_estimate(self):
        data = getData(1)
        my_biogeme = BIOGEME(data, self.get_dict_of_expressions(), bootstrap_samples=10)
        results = my_biogeme.estimate(run_bootstrap=True)
        self.assertAlmostEqual(results.final_log_likelihood, 0, 5)

    def test_estimate_panel(self):
        likelihood = self.get_dict_of_expressions_for_panel()
        panel_biogeme = BIOGEME(
            getPanelData(1),
            likelihood,
            bootstrap_samples=10,
            save_iterations=False,
        )
        results = panel_biogeme.estimate(run_bootstrap=True)
        self.assertAlmostEqual(results.final_log_likelihood, 0, 5)

    def test_estimate_scipy(self):
        data = getData(1)
        my_biogeme = BIOGEME(
            data,
            self.get_dict_of_expressions(),
            bootstrap_samples=10,
            optimization_algorithm='scipy',
            save_iterations=True,
            generate_yaml=True,
        )
        my_biogeme.model_name = 'test_scipy'
        results = my_biogeme.estimate()
        self.assertAlmostEqual(results.final_log_likelihood, 0, 5)
        # We try to recycle, while there is no yaml file yet.
        results = my_biogeme.estimate(recycle=True)
        # We estimate the model twice to generate two pickle .py.
        self.assertAlmostEqual(results.final_log_likelihood, 0, 5)

    def test_estimate_tr_newton(self):
        data = getData(1)
        my_biogeme = BIOGEME(
            data,
            self.get_dict_of_expressions(),
            bootstrap_samples=10,
            optimization_algorithm='TR-newton',
            save_iterations=False,
            generate_yaml=False,
        )
        results = my_biogeme.estimate()
        self.assertAlmostEqual(results.final_log_likelihood, 0, 5)

    def test_estimate_tr_bfgs(self):
        data = getData(1)
        my_biogeme = BIOGEME(
            data,
            self.get_dict_of_expressions(),
            bootstrap_samples=10,
            optimization_algorithm='TR-BFGS',
            save_iterations=False,
            generate_yaml=False,
        )
        results = my_biogeme.estimate()
        self.assertAlmostEqual(results.final_log_likelihood, 0, 5)

    def test_estimate_ls_newton(self):
        data = getData(1)
        my_biogeme = BIOGEME(
            data,
            self.get_dict_of_expressions(),
            bootstrap_samples=10,
            optimization_algorithm='LS-newton',
            save_iterations=False,
            generate_yaml=False,
        )
        results = my_biogeme.estimate()
        self.assertAlmostEqual(results.final_log_likelihood, 0, 5)

    def test_estimate_ls_bfgs(self):

        data = getData(1)
        my_biogeme = BIOGEME(
            data,
            self.get_dict_of_expressions(),
            bootstrap_samples=10,
            optimization_algorithm='LS-BFGS',
            save_iterations=False,
            generate_yaml=False,
        )
        results = my_biogeme.estimate()
        self.assertAlmostEqual(results.final_log_likelihood, 0, 5)

    def test_algo_error(self):
        data = getData(1)
        with self.assertRaises(BiogemeError):
            _ = BIOGEME(
                data,
                self.get_dict_of_expressions(),
                bootstrap_samples=10,
                optimization_algorithm='any_algo',
                save_iterations=False,
                generate_yaml=False,
            )

        with self.assertRaises(BiogemeError):
            _ = BIOGEME(
                data,
                self.get_dict_of_expressions(),
                bootstrap_samples=10,
                wrong_parameter_name='any_algo',
                save_iterations=False,
                generate_yaml=False,
            )

    def test_trivial_loglikelihood(self):
        aBiogeme = BIOGEME(getData(1), {'log_like': Numeric(0)})
        with self.assertRaises(BiogemeError):
            _ = aBiogeme.estimate()

    def test_quickEstimate(self):
        data = getData(1)
        my_biogeme = BIOGEME(
            data,
            self.get_dict_of_expressions(),
            save_iterations=False,
            generate_yaml=False,
            generate_html=False,
        )
        my_biogeme.model_name = 'simple_example'
        results = my_biogeme.quick_estimate()
        self.assertAlmostEqual(results.final_log_likelihood, 0, 5)

    def test_simulate(self):
        my_biogeme = self.get_biogeme_instance()
        results = my_biogeme.estimate()
        with self.assertRaises(BiogemeError):
            _ = my_biogeme.simulate(the_beta_values=None)
        my_biosim = self.get_biogeme_instance()
        s = my_biosim.simulate(results.get_beta_values())
        self.assertAlmostEqual(s.loc[0, 'log_like'], 0, 3)

        the_betas = results.get_beta_values()
        the_betas['any_beta'] = 0.1
        s = my_biosim.simulate(the_betas)
        self.assertAlmostEqual(s.loc[0, 'log_like'], 0, 3)

    def test_simulate_wrong(self):
        my_biogeme = self.get_biogeme_instance()
        with self.assertRaises(BiogemeError):
            _ = my_biogeme.simulate('wrong_object')

    def test_changeInitValues(self):
        my_biogeme = self.get_biogeme_instance()
        my_biogeme.change_init_values({'beta2': -100, 'beta1': 3.14156})
        self.assertListEqual(
            my_biogeme.model_elements.expressions_registry.list_of_free_betas_init_values,
            [3.14156, -100],
        )

    def test_confidenceIntervals(self):
        data = getData(1)
        my_biogeme = BIOGEME(
            data,
            self.get_dict_of_expressions(),
            save_iterations=False,
            generate_yaml=False,
            generate_html=False,
            bootstrap_samples=100,
        )
        results = my_biogeme.estimate(run_bootstrap=True)
        draws_from_betas = results.get_betas_for_sensitivity_analysis()
        my_biosim = BIOGEME(
            data,
            self.get_dict_of_expressions(),
            save_iterations=False,
            generate_yaml=False,
            generate_html=False,
        )
        s = my_biosim.simulate(results.get_beta_values())
        left, right = my_biosim.confidence_intervals(draws_from_betas)
        self.assertLessEqual(left.loc[0, 'log_like'], s.loc[0, 'log_like'])
        self.assertGreaterEqual(right.loc[0, 'log_like'], s.loc[0, 'log_like'])

    def test_validate(self):
        my_data = getData(1)
        my_biogeme = self.get_biogeme_instance()
        my_biogeme.algorithm_name = 'simple_bounds'
        results = my_biogeme.estimate()
        validation_results: list[ValidationResult] = my_biogeme.validate(
            results, slices=5
        )
        self.assertAlmostEqual(
            validation_results[0]
            .simulated_values['log_like [validation fold 1]']
            .sum(),
            0,
            3,
        )
        self.assertAlmostEqual(
            validation_results[1]
            .simulated_values['log_like [validation fold 2]']
            .sum(),
            0,
            3,
        )

    def test_print(self):
        my_biogeme = self.get_biogeme_instance()
        result = str(my_biogeme)[0:20]
        expected_result = 'biogeme_model_defaul'
        self.assertEqual(result, expected_result)

    def test_files(self):
        data = getData(1)
        my_biogeme = BIOGEME(
            data,
            self.get_dict_of_expressions(),
            save_iterations=False,
            generate_yaml=False,
            generate_html=True,
            bootstrap_samples=100,
        )
        my_biogeme.model_name = 'name_for_file'
        my_biogeme.estimate()
        result = files_of_type(
            name=my_biogeme.model_name, extension='html', all_files=False
        )
        expected_result = 'name_for_file.html'
        self.assertIn(expected_result, result, f"{expected_result} is not in the list")
        result = files_of_type('html', name=my_biogeme.model_name, all_files=True)
        self.assertGreaterEqual(len(result), 1)

    def test_from_configuration(self):
        beta_1 = Beta('beta_1', 0, None, None, 0)
        beta_2 = Beta('beta_2', 0, None, None, 0)
        catalog = Catalog.from_dict(
            catalog_name='the_catalog',
            dict_of_expressions={
                'beta_1': beta_1,
                'beta_2': beta_2,
            },
        )
        # Valid constructor
        config_id = 'the_catalog:beta_2'
        the_biogeme = BIOGEME.from_configuration(
            config_id=config_id,
            multiple_expression=catalog,
            database=getData(1),
        )

        # Invalid constructor: typo
        config_id = 'the_catalog:bta_2'
        with self.assertRaises(BiogemeError):
            _ = BIOGEME.from_configuration(
                config_id=config_id,
                multiple_expression=catalog,
                database=getData(1),
            )

        # Invalid constructor: wrong id
        config_id = 'wrong_id'
        with self.assertRaises(BiogemeError):
            _ = BIOGEME.from_configuration(
                config_id=config_id,
                multiple_expression=catalog,
                database=getData(1),
            )

    def test_report_array(self):
        my_biogeme = self.get_biogeme_instance()
        array = np.array([12, 45])
        result = my_biogeme.report_array(array)
        expected_result = 'beta1=12, beta2=45'
        self.assertEqual(result, expected_result)
        result = my_biogeme.report_array(array, with_names=False)
        expected_result = '12, 45'
        self.assertEqual(result, expected_result)

    def test_calculate_second_derivatives(self):

        data = getData(1)

        biogeme_0 = BIOGEME(data, self.get_dict_of_expressions())
        self.assertEqual(
            biogeme_0.second_derivatives_mode, SecondDerivativesMode.ANALYTICAL
        )

        biogeme_1 = BIOGEME(
            data, self.get_dict_of_expressions(), calculating_second_derivatives='never'
        )
        self.assertEqual(biogeme_1.second_derivatives_mode, SecondDerivativesMode.NEVER)

        biogeme_2 = BIOGEME(
            data,
            self.get_dict_of_expressions(),
            calculating_second_derivatives='finite_differences',
        )
        self.assertEqual(
            biogeme_2.second_derivatives_mode, SecondDerivativesMode.FINITE_DIFFERENCES
        )

        biogeme_3 = BIOGEME(
            data,
            self.get_dict_of_expressions(),
            calculating_second_derivatives='analytical',
        )
        self.assertEqual(
            biogeme_3.second_derivatives_mode, SecondDerivativesMode.ANALYTICAL
        )

        with self.assertRaises(BiogemeError):
            _ = BIOGEME(
                data,
                self.get_dict_of_expressions(),
                calculating_second_derivatives='xxx',
            )


if __name__ == '__main__':

    unittest.main()



"""Object containing the output of numerical expression  calculation"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Protocol, runtime_checkable, Sequence, TypeVar

import numpy as np

GenericType = TypeVar('GenericType')


def convert_to_dict(
    the_sequence: Sequence[GenericType], the_map: dict[str, int]
) -> dict[str, GenericType]:
    """Convert an indexed sequence into a dict mapping a name to the elements.

    :param the_sequence: sequence of element to convert
    :param the_map: dict mapping the names with their index in the sequence.
    :return: the sequence converted into dict
    """
    # Check for any index out of range errors before creating the dictionary
    for name, index in the_map.items():
        if index >= len(the_sequence) or index < 0:
            raise IndexError(
                f"Index {index} for variable '{name}' is out of range for sequence of length {len(the_sequence)}."
            )

    # Use dictionary comprehension to create the result
    result = {name: the_sequence[index] for name, index in the_map.items()}
    return result


@dataclass
class FunctionOutput:
    """Output of a function calculation"""

    function: float
    gradient: np.ndarray | None = None
    hessian: np.ndarray | None = None
    bhhh: np.ndarray | None = None


@dataclass
class DisaggregateFunctionOutput:
    """Output of a function calculation"""

    functions: np.ndarray
    gradients: np.ndarray | None = None
    hessians: np.ndarray | None = None
    bhhhs: np.ndarray | None = None

    def __len__(self):
        return len(self.functions)

    def unique_entry(self) -> FunctionOutput | None:
        """When there is only one entry, we generate the BiogemeFunctionOutput object"""
        if len(self) == 1:
            return FunctionOutput(
                function=float(self.functions[0]),
                gradient=self.gradients[0] if self.gradients is not None else None,
                hessian=self.hessians[0] if self.hessians is not None else None,
                bhhh=self.bhhhs[0] if self.bhhhs is not None else None,
            )
        return None


@runtime_checkable
class NamedFunctionResult(Protocol):
    """Protocol defining the interface for named function outputs."""

    function: float
    gradient: dict[str, float] | None
    hessian: dict[str, dict[str, float]] | None
    bhhh: dict[str, dict[str, float]] | None


class NamedOutputMixin:
    """Mixin providing common methods for named outputs."""

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}: function={self.function}, "
            f"gradient={self.gradient}, hessian={self.hessian}, bhhh={self.bhhh}>"
        )


@dataclass
class NamedFunctionOutput(NamedOutputMixin, NamedFunctionResult):
    """Output of a function calculation, with names of variables"""

    function_output: FunctionOutput
    mapping: dict[str, int]

    function: float = None
    gradient: dict[str, float] | None = None
    hessian: dict[str, dict[str, float]] | None = None
    bhhh: dict[str, dict[str, float]] | None = None

    def __post_init__(self):
        self.function = self.function_output.function
        self.gradient = (
            None
            if self.function_output.gradient is None
            else convert_to_dict(self.function_output.gradient, self.mapping)
        )
        self.hessian = (
            None
            if self.function_output.hessian is None
            else convert_to_dict(
                [
                    convert_to_dict(row, self.mapping)
                    for row in self.function_output.hessian
                ],
                self.mapping,
            )
        )
        self.bhhh = (
            None
            if self.function_output.bhhh is None
            else convert_to_dict(
                [
                    convert_to_dict(row, self.mapping)
                    for row in self.function_output.bhhh
                ],
                self.mapping,
            )
        )



import unittest
from unittest.mock import MagicMock, patch, Mock

import numpy as np
import pandas as pd

from biogeme.database import Database
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Numeric, Beta, Expression
from biogeme.mdcev.mdcev import Mdcev


# A concrete subclass of Mdcev for testing purposes
class ConcreteMdcev(Mdcev):
    def transformed_utility(self, the_id, the_consumption, epsilon=None) -> Expression:
        return Numeric(the_id)

    def calculate_log_determinant_one_alternative(
        self, the_id, consumption
    ) -> Expression:
        return Numeric(the_id)

    def calculate_inverse_of_determinant_one_alternative(
        self, the_id, consumption
    ) -> Expression:
        return Numeric(the_id)  # Implement abstract method

    def _list_of_expressions(self) -> list[Expression]:
        """Extract the list of expressions involved in the model"""
        return [Numeric(1), Numeric(2)]

    def utility_one_alternative(
        self,
        the_id: int,
        the_consumption: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        return 10

    def utility_expression_one_alternative(
        self,
        the_id: int,
        the_consumption: Expression,
        unscaled_epsilon: Expression,
    ) -> Expression:
        """Utility expression. Used only for code validation."""
        beta = Beta('beta', 10, None, None, 0)
        return beta

    def derivative_utility_one_alternative(
        self,
        the_id: int,
        the_consumption: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Used in the optimization problem solved for forecasting tp calculate the dual variable."""
        return 1

    def optimal_consumption_one_alternative(
        self,
        the_id: int,
        dual_variable: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Analytical calculation of the optimal consumption if the dual variable is known."""
        return dual_variable

    def lower_bound_dual_variable(
        self,
        chosen_alternatives: set[int],
        one_observation: Database,
        epsilon: np.ndarray,
    ) -> float:
        return 0.0


class TestConcreteMdcev(unittest.TestCase):

    def setUp(self):
        # Set up the common attributes used across many tests
        self.model_name = 'TestModel'
        self.baseline_utilities = {1: Mock(spec=Expression), 2: Mock(spec=Expression)}
        self.gamma_parameters = {1: Mock(spec=Expression), 2: None}
        self.alpha_parameters = {1: Mock(spec=Expression), 2: Mock(spec=Expression)}
        self.scale_parameter = Mock(spec=Expression)
        self.weights = Mock(spec=Expression)

        # Create a concrete instance of Mdcev for testing
        self.instance = ConcreteMdcev(
            model_name=self.model_name,
            baseline_utilities=self.baseline_utilities,
            gamma_parameters=self.gamma_parameters,
            alpha_parameters=self.alpha_parameters,
            scale_parameter=self.scale_parameter,
            weights=self.weights,
        )

    def test_initialization(self):
        # Test correct initialization
        self.assertEqual(self.instance.model_name, self.model_name)
        self.assertEqual(self.instance.baseline_utilities, self.baseline_utilities)
        self.assertEqual(self.instance.gamma_parameters, self.gamma_parameters)
        self.assertEqual(self.instance.alpha_parameters, self.alpha_parameters)
        self.assertEqual(self.instance.scale_parameter, self.scale_parameter)
        self.assertEqual(self.instance.weights, self.weights)

    def test_calculate_utilities(self):
        # Test calculation of utilities based on a mock consumption dict
        consumption = {1: Mock(spec=Expression), 2: Mock(spec=Expression)}
        with patch.object(
            self.instance, 'transformed_utility', return_value=Mock(spec=Expression)
        ) as mocked_transformed_utility:
            result = self.instance.calculate_utilities(consumption)
            self.assertEqual(set(result.keys()), set(consumption.keys()))
            mocked_transformed_utility.assert_called()

    def test_calculate_log_determinant_entries(self):
        # Test calculation of log determinant entries
        consumption = {1: Mock(spec=Expression), 2: Mock(spec=Expression)}
        with patch.object(
            self.instance,
            'calculate_log_determinant_one_alternative',
            return_value=Mock(spec=Expression),
        ) as mocked_method:
            result = self.instance.calculate_log_determinant_entries(consumption)
            self.assertEqual(set(result.keys()), set(consumption.keys()))
            self.assertEqual(mocked_method.call_count, len(consumption))

    def test_error_on_duplicate_keys_in_baseline_utilities(self):
        # Test that duplicate keys in baseline_utilities raise an error
        with self.assertRaises(BiogemeError):
            ConcreteMdcev(
                model_name=self.model_name,
                baseline_utilities={1: Mock(), 1: Mock()},  # Duplicate keys
                gamma_parameters=self.gamma_parameters,
            )

    def test_abstract_method_implementation_error(self):
        # Ensure that instantiation without implementing abstract methods raises TypeError
        with self.assertRaises(TypeError):
            Mdcev()  # Attempt to instantiate the abstract base class without a concrete implementation

    def test_forecast_method(self):
        # Testing the forecast method under controlled conditions
        self.instance.alternatives = {1, 2}  # Assuming two alternatives

        # Assuming two observations
        database = Mock(spec=Database)
        database.data = pd.DataFrame({'Column1': [1, 2], 'Column2': [1, 2]})
        number_of_draws = 10
        epsilons = [
            np.random.gumbel(
                loc=0,
                scale=1,
                size=(number_of_draws, 2),
            )
            for _ in range(2)
        ]

        total_budget = 100

        brute_force = False
        tolerance_dual = 1e-4
        tolerance_budget = 1e-4

        with patch.object(
            self.instance, 'forecast_bisection_one_draw', return_value={1: 50, 2: 50}
        ) as mocked_forecast:
            results = self.instance.forecast(
                database,
                total_budget,
                epsilons,
                brute_force,
                tolerance_dual,
                tolerance_budget,
            )
            self.assertIsInstance(results, list)
            self.assertEqual(len(results), 2)
            self.assertEqual(len(results[0]), number_of_draws)
            mocked_forecast.assert_called()

    def test_validation_method(self):
        # Testing the validation method
        one_row = Mock(spec=Database)
        with patch.object(
            self.instance, 'validation_one_alternative', return_value=[]
        ) as mocked_validation:
            result = self.instance.validation(one_row)
            self.assertIsInstance(result, list)
            mocked_validation.assert_called()

    def test_error_handling(self):
        # Test error handling in method that requires specific inputs
        database = Database('test', pd.DataFrame({'Column1': [1], 'Column2': [1]}))
        with self.assertRaises(KeyError):
            self.instance.calculate_baseline_utility(
                alternative_id=3, one_observation=database
            )  # Non-existent alternative ID


class TestUpdateParametersInExpressions(unittest.TestCase):
    def setUp(self):

        baseline_utilities = {
            1: Beta('beta1', 0, None, None, 0),
            2: Beta('beta2', 0, None, None, 0),
        }
        gamma_parameters = {
            1: Beta('gamma1', 0, None, None, 0),
            2: None,
        }
        alpha_parameters = {
            1: Beta('alpha1', 0, None, None, 0),
            2: Beta('alpha2', 0, None, None, 0),
        }
        scale_parameter = Beta('scale', 0, None, None, 0)
        weights = Numeric(1)

        self.model = ConcreteMdcev(
            model_name='test',
            baseline_utilities=baseline_utilities,
            gamma_parameters=gamma_parameters,
            alpha_parameters=alpha_parameters,
            scale_parameter=scale_parameter,
            weights=weights,
        )

    def test_update_parameters(self):
        # Mock estimation_results to simulate get_beta_values
        # alpha1 is omitted on purpose.

        estimation_results = {
            'beta1': 1,
            'beta2': 2,
            'gamma1': 3,
            'alpha2': 4,
            'scale': 5,
            'any_parameter': 6,
        }
        self.model._estimation_results = MagicMock()
        self.model._estimation_results.get_beta_values.return_value = estimation_results
        self.model._update_parameters_in_expressions()

        self.assertEqual(
            self.model.baseline_utilities[1].get_value(),
            estimation_results['beta1'],
        )
        self.assertEqual(
            self.model.baseline_utilities[2].get_value(),
            estimation_results['beta2'],
        )
        self.assertEqual(
            self.model.gamma_parameters[1].get_value(),
            estimation_results['gamma1'],
        )
        self.assertEqual(
            self.model.alpha_parameters[1].get_value(),
            0,
        )

        self.assertEqual(
            self.model.alpha_parameters[2].get_value(),
            estimation_results['alpha2'],
        )
        self.assertEqual(
            self.model.scale_parameter.get_value(),
            estimation_results['scale'],
        )


if __name__ == "__main__":
    unittest.main()



"""
Old estimation result object.

Michel Bierlaire
Mon Oct 7 13:53:50 2024
"""

from biogeme.exceptions import BiogemeError

text = """
The old and unstructured `bioResults` object is now obsolete. Biogeme has introduced a new, more streamlined object 
called `EstimationResults`, which you can access as follows:

    from biogeme.results_processing.estimation_results import EstimationResults

Transition from Pickle to YAML

Biogeme has discontinued using the `pickle` format for saving estimation results due to its lack of transparency 
and inconsistencies across different software versions. Instead, Biogeme now saves results in a YAML format, 
which is more transparent and readable.

Converting Pickle Files to YAML

If you have an existing pickle file, you can easily convert it to the YAML format with the following command:

    from biogeme.results_processing import pickle_to_yaml
    pickle_to_yaml(pickle_filename='my_model.pickle', yaml_filename='my_model.yaml')

Accessing Estimation Results from Pickle or YAML

You can create an `EstimationResults` object directly from either a pickle file or a YAML file as follows:

    from biogeme.results_processing import EstimationResults
    results = EstimationResults.from_pickle_file(filename='my_model.pickle')
    results = EstimationResults.from_yaml_file(filename='my_model.yaml')
"""


class bioResults:
    """Class managing the estimation results"""

    def __init__(self, *args, **kwargs):
        """Constructor"""

        raise BiogemeError(text)



from __future__ import annotations

import logging
from typing import NamedTuple, TYPE_CHECKING

import numpy as np
from tabulate import tabulate

from biogeme.floating_point import SQRT_EPS
from biogeme.function_output import FunctionOutput, NamedFunctionOutput

if TYPE_CHECKING:
    from biogeme.calculator import CallableExpression

logger = logging.getLogger(__name__)


class CheckDerivativesResults(NamedTuple):
    function: float
    analytical_gradient: np.ndarray
    analytical_hessian: np.ndarray
    finite_differences_gradient: np.ndarray
    finite_differences_hessian: np.ndarray
    errors_gradient: np.ndarray
    errors_hessian: np.ndarray


def findiff_g(the_function: CallableExpression, x: np.ndarray) -> np.ndarray:
    """Calculates the gradient of a function :math:`f` using finite differences

    :param the_function: A function object that takes a vector as an
                        argument, and returns a tuple. The first
                        element of the tuple is the value of the
                        function :math:`f`. The other elements are not
                        used.

    :param x: argument of the function

    :return: numpy vector, same dimension as x, containing the gradient
       calculated by finite differences.
    """
    x = x.astype(float)
    tau = SQRT_EPS
    n = len(x)
    g = np.zeros(n)
    f = the_function(x, gradient=False, hessian=False, bhhh=False).function
    for i in range(n):
        xi = x.item(i)
        xp = x.copy()
        if abs(xi) >= 1:
            s = tau * xi
        elif xi >= 0:
            s = tau
        else:
            s = -tau
        xp[i] = xi + s
        fp = the_function(xp, gradient=False, hessian=False, bhhh=False).function
        g[i] = (fp - f) / s
    return g


def findiff_h(
    the_function: CallableExpression,
    x: np.ndarray,
) -> np.ndarray:
    """Calculates the hessian of a function :math:`f` using finite differences


    :param the_function: A function object that takes a vector as an
                        argument, and returns a tuple. The first
                        element of the tuple is the value of the
                        function :math:`f`, and the second is the
                        gradient of the function.  The other elements
                        are not used.

    :param x: argument of the function
    :return: numpy matrix containing the hessian calculated by
             finite differences.
    """
    tau = SQRT_EPS
    n = len(x)
    h = np.zeros((n, n))
    the_function_output: FunctionOutput | NamedFunctionOutput = the_function(
        x, gradient=True, hessian=False, bhhh=False
    )
    if isinstance(the_function_output, NamedFunctionOutput):
        the_function_output = the_function_output.function_output
    g = the_function_output.gradient
    eye = np.eye(n, n)
    for i in range(n):
        xi = x.item(i)
        if abs(xi) >= 1:
            s = tau * xi
        elif xi >= 0:
            s = tau
        else:
            s = -tau
        ei = eye[i]
        the_function_output: FunctionOutput | NamedFunctionOutput = the_function(
            x + s * ei, gradient=True, hessian=False, bhhh=False
        )
        if isinstance(the_function_output, NamedFunctionOutput):
            the_function_output = the_function_output.function_output
        gp = the_function_output.gradient
        h[:, i] = (gp - g).flatten() / s
    return h


def check_derivatives(
    the_function: CallableExpression,
    x: np.ndarray,
    names: list[str] | None = None,
    logg: bool | None = False,
) -> CheckDerivativesResults:
    """Verifies the analytical derivatives of a function by comparing
    them with finite difference approximations.

    :param the_function: A function object that takes a vector as an argument,
        and returns a tuple:

        - The first element of the tuple is the value of the
          function :math:`f`,
        - the second is the gradient of the function,
        - the third is the hessian.

    :param x: arguments of the function

    :param names: the names of the entries of x (for reporting).

    :param logg: if True, messages will be displayed.

    :return: tuple f, g, h, gdiff, hdiff where

          - f is the value of the function at x,
          - g is the analytical gradient,
          - h is the analytical hessian,
          - gdiff is the difference between the analytical gradient
            and the finite difference approximation
          - hdiff is the difference between the analytical hessian
            and the finite difference approximation

    """
    x = np.array(x, dtype=float)
    the_function_output: FunctionOutput | NamedFunctionOutput = the_function(
        x, gradient=True, hessian=True, bhhh=False
    )
    if isinstance(the_function_output, NamedFunctionOutput):
        the_function_output = the_function_output.function_output
    g_num = findiff_g(the_function, x)
    gdiff = the_function_output.gradient - g_num
    if logg:
        headers = ['x', 'Gradient', 'FinDiff', 'Difference']

        if names is None:
            names = [f'x[{i}]' for i in range(len(x))]
        rows = [
            [
                f'{names[k]}',
                f'{the_function_output.gradient[k]:+E}',
                f'{g_num[k]:+E}',
                f'{v:+E}',
            ]
            for k, v in enumerate(gdiff)
        ]
        logger.info('Comparing first derivatives')
        logger.info(tabulate(rows, headers=headers, tablefmt='plain'))

    h_num = findiff_h(the_function, x)
    hdiff = the_function_output.hessian - h_num
    if logg:
        headers = ['Row', 'Col', 'Hessian', 'FinDiff', 'Difference']
        rows = [
            [
                names[row],
                names[col],
                f'{the_function_output.hessian[row, col]:+E}',
                f'{h_num[row, col]:+E}',
                f'{hdiff[row, col]:+E}',
            ]
            for col in range(len(hdiff))
            for row in range(len(hdiff))
        ]
        logger.info('Comparing second derivatives')
        logger.info(tabulate(rows, headers=headers, tablefmt='plain'))

    return CheckDerivativesResults(
        function=the_function_output.function,
        analytical_gradient=the_function_output.gradient,
        analytical_hessian=the_function_output.hessian,
        finite_differences_gradient=g_num,
        finite_differences_hessian=h_num,
        errors_gradient=gdiff,
        errors_hessian=hdiff,
    )



"""
Generates an HTML output

Michel Bierlaire
Wed Oct 2 16:10:54 2024
"""

import logging
import os
from datetime import datetime

import biogeme.version as version
import numpy as np
from biogeme.exceptions import BiogemeError

from .estimation_results import (
    EstimateVarianceCovariance,
    EstimationResults,
    calc_p_value,
    calculates_correlation_matrix,
)
from ..parameters import Parameters

logger = logging.getLogger(__name__)


def format_real_number(value: float) -> str:
    """Format a real number to be included in the HTML table"""

    formatted_value = f'{value:.3g}'
    return formatted_value


def get_html_header(estimation_results: EstimationResults) -> str:
    """Prepare the header for the HTML file, containing comments and the
    version of Biogeme.

    :return: string containing the header.
    """
    html = ''
    html += '<html>\n'
    html += '<head>\n'
    html += '<script src="http://transp-or.epfl.ch/biogeme/sorttable.js">' '</script>\n'
    html += '<meta http-equiv="Content-Type" content="text/html; ' 'charset=utf-8" />\n'
    if estimation_results.raw_estimation_results is None:
        html += '<p>No estimation result is available.</p>'
        return html
    html += (
        f'<title>{estimation_results.raw_estimation_results.model_name} - Report from '
        f'biogeme {version.get_version()} '
        f'[{version.versionDate}]</title>\n'
    )
    html += (
        '<meta name="keywords" content="biogeme, discrete choice, ' 'random utility">\n'
    )
    html += (
        f'<meta name="description" content="Report from '
        f'biogeme {version.get_version()} '
        f'[{version.versionDate}]">\n'
    )
    html += '<meta name="author" content="{bv.author}">\n'
    html += '<style type=text/css>\n'
    html += '.biostyle\n'
    html += '	{font-size:10.0pt;\n'
    html += '	font-weight:400;\n'
    html += '	font-style:normal;\n'
    html += '	font-family:Courier;}\n'
    html += '.boundstyle\n'
    html += '	{font-size:10.0pt;\n'
    html += '	font-weight:400;\n'
    html += '	font-style:normal;\n'
    html += '	font-family:Courier;\n'
    html += '        color:red}\n'
    html += '</style>\n'
    html += '</head>\n'
    html += '<body bgcolor="#ffffff">\n'
    return html


def get_html_footer() -> str:
    """Prepare the footer for the HTML file, containing comments and the
    version of Biogeme.
    """
    html = '</body>\n</html>'
    return html


def get_html_preamble(estimation_results: EstimationResults, file_name: str) -> str:
    """Generates the first part of the HTML, with the preamble information.

    :param estimation_results: estimation results
    :param file_name: name of the HTML file (used only for reporting)
    :return: HTML code
    """
    now = datetime.now()
    html = version.get_html()
    html += f'<p>This file has automatically been generated on {now}</p>\n'
    if estimation_results.raw_estimation_results is None:
        html += '<p>No estimation result is available.</p>'
        return html
    html += '<table>\n'
    html += (
        f'<tr class=biostyle><td align=right>'
        f'<strong>Report file</strong>:	</td>'
        f'<td>{file_name}</td></tr>\n'
    )
    html += (
        f'<tr class=biostyle><td align=right>'
        f'<strong>Database name</strong>:	</td>'
        f'<td>{estimation_results.raw_estimation_results.data_name}</td></tr>\n'
    )
    html += '</table>\n'

    if not estimation_results.algorithm_has_converged:
        html += '<h2>Algorithm failed to converge</h2>\n'
        html += (
            '<p>It seems that the optimization algorithm did not converge. '
            'Therefore, the results below do not correspond to the maximum '
            'likelihood estimator. Check the specification of the model, '
            'or the criteria for convergence of the algorithm. </p>'
        )
    identification_threshold = Parameters().get_value('identification_threshold')
    try:
        if np.abs(estimation_results.smallest_eigenvalue) <= identification_threshold:
            html += '<h2>Warning: identification issue</h2>\n'
            html += (
                f'<p>The second derivatives matrix is close to singularity. '
                f'The smallest eigenvalue is '
                f'{np.abs(estimation_results.smallest_eigenvalue):.3g}. This warning is '
                f'triggered when it is smaller than the parameter '
                f'<code>identification_threshold</code>='
                f'{identification_threshold}.</p>'
                f'<p>Variables involved:'
            )
            html += '<table>'
            for i, ev in enumerate(estimation_results.smallest_eigenvector):
                if np.abs(ev) > identification_threshold:
                    html += (
                        f'<tr><td>{ev:.3g}</td>'
                        f'<td> *</td>'
                        f'<td> {estimation_results.beta_names[i]}</td></tr>\n'
                    )
            html += '</table>'
            html += '</p>\n'
    except BiogemeError:
        html += '<h2>Warning: second derivatives matrix not available.</h2>\n'
        html += '<p>The second derivatives matrix has not been calculated. The statistics requiring it are not generated.</p>'

    if estimation_results.user_notes is not None:
        # User notes
        html += (
            f'<blockquote style="border: 2px solid #666; '
            f'padding: 10px; background-color:'
            f' #ccc;">{estimation_results.user_notes}</blockquote>'
        )
    return html


def get_html_general_statistics(estimation_results: EstimationResults) -> str:
    """Get the general statistics coded in HTML

    :return: HTML code
    """

    html = '<table border="0">\n'
    statistics = estimation_results.get_general_statistics()
    for description, value in statistics.items():
        if value is not None:
            html += (
                f'<tr class=biostyle><td align=right >'
                f'<strong>{description}</strong>: </td> '
                f'<td>{value}</td></tr>\n'
            )
    for key, value in estimation_results.optimization_messages.items():
        if key in (
            'Relative projected gradient',
            'Relative change',
            'Relative gradient',
        ):
            html += (
                f'<tr class=biostyle><td align=right >'
                f'<strong>{key}</strong>: </td> '
                f'<td>{value:.7g}</td></tr>\n'
            )
        else:
            html += (
                f'<tr class=biostyle><td align=right >'
                f'<strong>{key}</strong>: </td> '
                f'<td>{value}</td></tr>\n'
            )

    html += '</table>\n'
    return html


def get_html_one_parameter(
    estimation_results: EstimationResults,
    parameter_index: int,
    variance_covariance_type: EstimateVarianceCovariance,
    parameter_number=None,
    parameter_name=None,
):
    """Generate the HTML code for one row of the table of the estimated parameters.

    :param estimation_results: estimation results.
    :param parameter_index: index of the parameter
    :param variance_covariance_type: type of variance-covariance estimate to be used.
    :param parameter_number: number of the parameter to report. If None, it is the index.
    :param parameter_name: name of the parameter to report. If None, taken from estimation results.
    :return: HTML code for the row
    """
    if parameter_index < 0 or parameter_index >= len(estimation_results.beta_names):
        error_msg = f'Invalid parameter index {parameter_index}. Valid range: 0- {len(estimation_results.beta_names)-1}'
        raise ValueError(error_msg)
    if parameter_number is None:
        parameter_number = parameter_index
    if parameter_name is None:
        parameter_name = estimation_results.beta_names[parameter_index]

    output = '<tr class=biostyle>'
    output += f'<td>{parameter_number}</td>'
    output += f'<td>{parameter_name}</td>'
    # Value
    value = estimation_results.get_parameter_value_from_index(
        parameter_index=parameter_index
    )
    output += f'<td>{format_real_number(value)}</td>'
    # std err
    std_err = estimation_results.get_parameter_std_err_from_index(
        parameter_index=parameter_index, estimate_var_covar=variance_covariance_type
    )
    output += f'<td>{format_real_number(std_err)}</td>'
    # t-test against 0
    t_test = estimation_results.get_parameter_t_test_from_index(
        parameter_index=parameter_index,
        estimate_var_covar=variance_covariance_type,
        target=0,
    )
    output += f'<td>{format_real_number(t_test)}</td>'
    # p-value against 0
    p_value = estimation_results.get_parameter_p_value_from_index(
        parameter_index=parameter_index,
        estimate_var_covar=variance_covariance_type,
        target=0,
    )
    output += f'<td>{format_real_number(p_value)}</td>'
    if estimation_results.is_bound_active(
        parameter_name=estimation_results.beta_names[parameter_index]
    ):
        output += '<td>Active bound</td>'
    output += '</tr>'
    return output


def get_html_estimated_parameters(
    estimation_results: EstimationResults,
    variance_covariance_type: EstimateVarianceCovariance,
    renaming_parameters: dict[str, str] | None = None,
    sort_by_name: bool = False,
) -> str:
    """Get the estimated parameters coded in HTML

    :param estimation_results: estimation results.
    :param variance_covariance_type: type of variance-covariance estimate to be used.
    :param renaming_parameters: a dict that suggests new names for some or all parameters.
    :param sort_by_name: if True, parameters are sorted alphabetically by name.
    :return: HTML code
    """
    if renaming_parameters is not None:
        # Verify that the renaming is well-defined.
        name_values = list(renaming_parameters.values())
        if len(name_values) != len(set(name_values)):
            warning_msg = (
                f'The new renaming assigns the same name for multiple parameters. It may not be the '
                f'desired action: {renaming_parameters}'
            )
            logger.warning(warning_msg)

    covar_header = str(variance_covariance_type)
    html = '<table border="1">\n'
    html += '<tr class=biostyle>'
    html += '<th>Id</th>'
    html += '<th>Name</th>'
    html += '<th>Value</th>'
    html += f'<th>{covar_header} std err.</th>'
    html += f'<th>{covar_header} t-stat.</th>'
    html += f'<th>{covar_header} p-value</th>'
    if estimation_results.is_any_bound_active():
        html += '<th></th>'
    html += '</tr>\n'

    rows = []
    for parameter_index, parameter_name in enumerate(estimation_results.beta_names):
        name = (
            renaming_parameters.get(parameter_name)
            if renaming_parameters is not None
            else parameter_name
        )
        row_html = (
            get_html_one_parameter(
                estimation_results=estimation_results,
                parameter_index=parameter_index,
                variance_covariance_type=variance_covariance_type,
                parameter_number=parameter_index,
                parameter_name=name,
            )
            + '\n'
        )
        rows.append((parameter_index, name, row_html))

    if sort_by_name:
        rows.sort(key=lambda x: x[1])

    for _, _, row_html in rows:
        html += row_html
    html += '</table>'
    return html


def get_html_one_pair_of_parameters(
    estimation_results: EstimationResults,
    first_parameter_index: int,
    second_parameter_index: int,
    variance_covariance_type: EstimateVarianceCovariance,
    first_parameter_name=None,
    second_parameter_name=None,
) -> str:
    """Generate one row of the HTML table of the correlation data for estimated parameters.

    :param estimation_results: estimation results.
    :param first_parameter_index: index of the first parameter
    :param second_parameter_index: index of the second parameter
    :param variance_covariance_type: type of variance-covariance estimate to be used.
    :param first_parameter_name: name of the parameter to report. If None, taken from estimation results.
    :param second_parameter_name: name of the parameter to report. If None, taken from estimation results.
    :return: one row of the table
    """
    if first_parameter_index < 0 or first_parameter_index >= len(
        estimation_results.beta_names
    ):
        error_msg = (
            f'Invalid parameter index {first_parameter_index}. Valid range: 0-'
            f' {len(estimation_results.beta_names)-1}'
        )
        raise ValueError(error_msg)
    if second_parameter_index < 0 or second_parameter_index >= len(
        estimation_results.beta_names
    ):
        error_msg = (
            f'Invalid parameter index {second_parameter_index}. Valid range: 0-'
            f' {len(estimation_results.beta_names)-1}'
        )
        raise ValueError(error_msg)

    if first_parameter_name is None:
        first_parameter_name = estimation_results.beta_names[first_parameter_index]
    if second_parameter_name is None:
        second_parameter_name = estimation_results.beta_names[second_parameter_index]

    covariance_matrix = estimation_results.get_variance_covariance_matrix(
        variance_covariance_type=variance_covariance_type
    )
    correlation_matrix = calculates_correlation_matrix(covariance=covariance_matrix)
    covariance = float(covariance_matrix[first_parameter_index, second_parameter_index])
    correlation = float(
        correlation_matrix[first_parameter_index, second_parameter_index]
    )
    t_test = estimation_results.calculate_test(
        first_parameter_index, second_parameter_index, covariance_matrix
    )
    p_value = calc_p_value(t_test)
    the_row = '<tr class=biostyle>'
    the_row += f'<td>{first_parameter_name}</td>'
    the_row += f'<td>{second_parameter_name}</td>'
    the_row += f'<td>{format_real_number(covariance)}</td>'
    the_row += f'<td>{format_real_number(correlation)}</td>'
    the_row += f'<td>{format_real_number(t_test)}</td>'
    the_row += f'<td>{format_real_number(p_value)}</td>'
    the_row += '</tr>'
    return the_row


def get_html_correlation_results(
    estimation_results: EstimationResults,
    variance_covariance_type: EstimateVarianceCovariance,
    involved_parameters: dict[str, str] | None = None,
) -> str:
    """Get the correlation results in an HTML format

    :param estimation_results: estimation results.
    :param variance_covariance_type: type of variance-covariance estimate to be used.
    :param involved_parameters: a dict that identifies the parameters to involve, as assign them with a name for the
        reporting.
    :return: HTML code
    """
    covar_header = str(variance_covariance_type)

    if involved_parameters is None:
        list_of_parameters = {
            index: name for index, name in enumerate(estimation_results.beta_names)
        }
    else:
        list_of_parameters = {
            estimation_results.get_parameter_index(orig_name): new_name
            for orig_name, new_name in involved_parameters.items()
        }

    html = '<table border="1">\n'
    html += '<tr class=biostyle>'
    html += '<th>Coefficient 1</th>'
    html += '<th>Coefficient 2</th>'
    html += f'<th>{covar_header} covariance</th>'
    html += f'<th>{covar_header} correlation</th>'
    html += f'<th>{covar_header} t-test</th>'
    html += f'<th>{covar_header} p-value</th>'
    html += '</tr>'
    for first_parameter_index, first_parameter_name in list_of_parameters.items():
        for second_parameter_index, second_parameter_name in list_of_parameters.items():
            if first_parameter_index > second_parameter_index:
                the_row = get_html_one_pair_of_parameters(
                    estimation_results=estimation_results,
                    first_parameter_index=first_parameter_index,
                    second_parameter_index=second_parameter_index,
                    variance_covariance_type=variance_covariance_type,
                    first_parameter_name=first_parameter_name,
                    second_parameter_name=second_parameter_name,
                )
                html += the_row + '\n'
    html += '</table>'
    return html


def get_html_condition_number(estimation_results: EstimationResults) -> str:
    """Report the smallest and largest eigenvalues, and the condition number.

    :param estimation_results: estimation results
    :return: HTML code
    """
    html = (
        f'<p>Smallest eigenvalue: '
        f'{estimation_results.smallest_eigenvalue:.6g}</p>\n'
    )
    html += (
        f'<p>Largest eigenvalue: ' f'{estimation_results.largest_eigenvalue:.6g}</p>\n'
    )
    html += f'<p>Condition number: ' f'{estimation_results.condition_number:.6g}</p>\n'
    return html


def generate_html_file(
    estimation_results: EstimationResults,
    filename: str,
    overwrite=False,
    variance_covariance_type: EstimateVarianceCovariance | None = None,
) -> None:
    """Generate an HTML file with the estimation results

    :param estimation_results: estimation results
    :param filename: name of the file
    :param overwrite: if True and the file exists, it is overwritten
    :param variance_covariance_type: select which type of variance-covariance matrix is used to generate the
        statistics. If None, the bootstrap one is used if available. If not available, the robust one.
    """
    if variance_covariance_type is None:
        variance_covariance_type = (
            estimation_results.get_default_variance_covariance_matrix()
        )
    if (
        variance_covariance_type == EstimateVarianceCovariance.BOOTSTRAP
        and estimation_results.bootstrap_time is None
    ):
        logger.warning(
            f'No bootstrap data is available. The robust variance-covariance matrix is used instead.'
        )
        variance_covariance_type = EstimateVarianceCovariance.ROBUST

    if not overwrite and os.path.exists(filename):
        raise FileExistsError(f"The file '{filename}' already exists.")

    with open(filename, 'w') as file:
        header = get_html_header(estimation_results=estimation_results)
        preamble = get_html_preamble(
            estimation_results=estimation_results, file_name=filename
        )
        general_statistics = get_html_general_statistics(
            estimation_results=estimation_results
        )
        parameters = get_html_estimated_parameters(
            estimation_results=estimation_results,
            sort_by_name=True,
            variance_covariance_type=variance_covariance_type,
        )
        correlation_results = get_html_correlation_results(
            estimation_results=estimation_results,
            variance_covariance_type=variance_covariance_type,
        )

        footer = get_html_footer()

        print(header, file=file)
        print(preamble, file=file)
        print('<h1>Estimation report</h1>', file=file)
        print(general_statistics, file=file)
        print('<h1>Estimated parameters</h1>', file=file)
        print(parameters, file=file)
        print('<h2>Correlation of coefficients</h2>', file=file)
        print(correlation_results, file=file)
        try:
            condition_number = get_html_condition_number(
                estimation_results=estimation_results
            )
            print(condition_number, file=file)
        except BiogemeError:
            ...
        print(footer, file=file)
    logger.info(f'File {filename} has been generated.')



import unittest
import subprocess
import os

from icecream import ic


class TestScripts(unittest.TestCase):

    def setUp(self):
        # Directory where the scripts are located
        self.script_dir = [
            './assisted',
            './indicators',
            './latent',
            './latentbis',
            './montecarlo',
            './programmers',
            './sampling',
            './swissmetro',
        ]

    def get_python_scripts(self):
        # Get all .py files in the specified directories
        return [
            (
                f,
                directory,
            )  # Join each directory with each file and include directory
            for directory in self.script_dir  # Iterate over each directory
            for f in os.listdir(directory)  # List files in the current directory
            if f.endswith('.py')  # Only Python files
        ]

    def test_run_scripts(self):
        # Get all Python scripts and their directories
        python_scripts = self.get_python_scripts()
        ic(python_scripts)
        for script_path, script_dir in python_scripts:
            with self.subTest(script=script_path):
                # Run the script as a subprocess in the correct directory (script_dir)
                result = subprocess.run(
                    ['python', script_path],
                    capture_output=True,
                    text=True,
                    cwd=script_dir,  # Set the working directory to the script's directory
                )

                # Check if the script ran successfully
                self.assertEqual(
                    result.returncode,
                    0,
                    f"Script {script_path} failed with output: {result.stdout}\nError: {result.stderr}",
                )


if __name__ == '__main__':
    unittest.main()



"""Functions to verify the validity of parameters

:author: Michel Bierlaire
:date: Thu Dec  1 16:22:34 2022

"""

from __future__ import annotations

import numbers
from typing import TYPE_CHECKING

import biogeme.optimization as opt
from biogeme.second_derivatives import SecondDerivativesMode

if TYPE_CHECKING:
    from biogeme.default_parameters import ParameterValue


def is_number(x: ParameterValue) -> tuple[bool, str | None]:
    """Return true if x is a number

    :param x: value of the parameter to check
    """
    if isinstance(x, numbers.Number):
        return True, None
    return False, 'Value must be a number'


def zero_one(x: ParameterValue) -> tuple[bool, str | None]:
    """Return true if x is between zero and one

    :param x: value of the parameter to check

    """
    check, msg = is_number(x)
    if not check:
        return check, msg
    if 0 <= x <= 1:
        return True, None
    return False, 'Value must be between zero and one'


def is_positive(x: ParameterValue) -> tuple[bool, str | None]:
    """Return true if x is positive

    :param x: value of the parameter to check
    """
    check, msg = is_number(x)
    if not check:
        return check, msg

    if x > 0:
        return True, None
    return False, 'Value must be positive'


def is_non_negative(x: ParameterValue) -> tuple[bool, str | None]:
    """Return true if x is non_negative

    :param x: value of the parameter to check

    """
    if x >= 0:
        return True, None
    return False, 'Value must be non negative'


def is_integer(x: ParameterValue) -> tuple[bool, str | None]:
    """Return true if x is integer

    :param x: value of the parameter to check
    """
    if isinstance(x, numbers.Integral):
        return True, None
    return False, 'Value must be an integer'


def check_algo_name(x: ParameterValue) -> tuple[bool, str | None]:
    """Return true if x is a valid algorithm name

    :param x: value of the parameter to check
    """
    if not isinstance(x, str):
        return False, f'Parameter must be a string: {x}'
    possibilities = ['automatic'] + list(opt.algorithms.keys())
    if x in possibilities:
        return True, None
    return False, f'Value must be in: {possibilities}'


def check_calculating_second_derivatives(x: ParameterValue) -> tuple[bool, str | None]:
    """Return true if x is a valid way to calculate the second derivatives

    :param x: value of the parameter to check
    """
    if not isinstance(x, str):
        return False, f'Parameter must be a string: {x}'
    possibilities = [value.value for value in SecondDerivativesMode]
    if x in possibilities:
        return True, None
    return False, f'Value must be in: {possibilities}'


def is_boolean(x: ParameterValue) -> tuple[bool, str | None]:
    """Return true if x is a boolean

    :param x: value of the parameter to check
    :type x: float
    """
    if isinstance(x, bool):
        return True, None
    return False, 'Value must be boolean'



def format_memory_size(num_bytes: float) -> str:
    """Format a number of bytes in human readable format

    :param num_bytes: number of bytes
    :return: string to be displayed
    """
    for unit in ['bytes', 'KB', 'MB', 'GB', 'TB']:
        if num_bytes < 1024:
            return f"{num_bytes:.2f} {unit}"
        num_bytes /= 1024
    return f'{num_bytes:.2f} PB'  # for very large numbers



import unittest

from biogeme.calculator import get_value_and_derivatives
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Beta, Elem, Numeric


class TestElemExpression(unittest.TestCase):

    def test_valid_key(self):
        expr = Elem({1: Numeric(10), 2: Numeric(20)}, Numeric(1))
        self.assertEqual(expr.get_value(), 10)

    def test_valid_negative_key(self):
        expr = Elem({-1: Numeric(5), 0: Numeric(15)}, Numeric(-1))
        self.assertEqual(expr.get_value(), 5)

    def test_invalid_key_raises(self):
        expr = Elem({1: Numeric(10)}, Numeric(3))
        with self.assertRaises(BiogemeError):
            expr.get_value()

    def test_derivative_with_respect_to_parameter(self):
        b1 = Beta('b1', 1.0, None, None, 0)
        b2 = Beta('b2', 2.0, None, None, 0)
        expr = Elem({0: b1 * b1, 1: b2 * b2}, Numeric(1))
        result = get_value_and_derivatives(expression=expr, numerically_safe=False)
        value = result.function
        gradient = list(result.gradient)
        expected_gradient = [0, 4]
        self.assertEqual(value, 4.0)
        self.assertListEqual(gradient, expected_gradient)

        expr = Elem({0: b1 * b1, 1: b2 * b2}, Numeric(2))
        result = get_value_and_derivatives(expression=expr, numerically_safe=False)
        value = result.function
        gradient = list(result.gradient)
        expected_gradient = [2, 0]
        self.assertEqual(value, 1.0)
        self.assertListEqual(gradient, expected_gradient)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from collections.abc import Iterable
from typing import TYPE_CHECKING

from .segmentation import Segmentation
from .segmentation_context import DiscreteSegmentationTuple

if TYPE_CHECKING:
    from biogeme.expressions import Beta


def segmented_beta(
    beta: Beta,
    segmentation_tuples: Iterable[DiscreteSegmentationTuple],
    prefix: str = 'segmented',
):
    """Obtain the segmented Beta from a unique function call

    :param beta: parameter to be segmented
    :param segmentation_tuples: characterization of the segmentations
    :param prefix: prefix to be used to generate the name of the
        segmented parameter
    :return: expression of the segmented Beta
    """
    the_segmentation = Segmentation(
        beta=beta, segmentation_tuples=segmentation_tuples, prefix=prefix
    )
    return the_segmentation.segmented_beta()



"""Decorator for the derivative calculation, so that the names of the variables are associated with their values.

:author: Michel Bierlaire
:date: Sat Apr 20 08:06:54 2024
"""

from functools import wraps

from biogeme.expressions import Expression
from biogeme.function_output import FunctionOutput, NamedFunctionOutput, convert_to_dict


def named_function_output(func):
    """
    Decorator to convert a FunctionOutput into a NamedFunctionOutput,
    using explicitly passed beta indices.

    The decorated function must now accept an `indices` keyword argument.
    """

    @wraps(func)
    def wrapper(self: Expression, *args, indices=None, **kwargs):
        if indices is None:
            raise ValueError(
                "Parameter 'indices' must be provided to use @named_function_output."
            )
        # Call the original function/method
        result: FunctionOutput = func(self, *args, **kwargs)
        if not isinstance(result, FunctionOutput):
            raise TypeError(
                f'Expected function {func.__name__} to return FunctionOutput, got {type(result).__name__} instead.'
            )

        the_result = NamedFunctionOutput(
            function=result.function,
            gradient=convert_to_dict(result.gradient, indices),
            hessian=convert_to_dict(
                [convert_to_dict(row, indices) for row in result.hessian],
                indices,
            ),
        )

        return the_result

    return wrapper



"""Split data into validation and estimation samples"""

from typing import NamedTuple
import logging

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


class EstimationValidationIndices(NamedTuple):
    estimation: pd.Index
    validation: pd.Index


def split(
    dataframe: pd.DataFrame, slices: int, groups: str | None = None
) -> list[EstimationValidationIndices]:
    """
    Splits a DataFrame into multiple training and validation index sets for cross-validation.

    This function returns a list of `EstimationValidationIndices` named tuples, each containing
    the indices for an estimation (training) set and a validation set. If a grouping column
    is specified, the split ensures that all entries with the same group ID remain in the
    same fold.

    :param dataframe: The full dataset to split.
    :param slices: The number of folds/slices. Must be >= 2.
    :param groups: Optional name of the column containing group identifiers.
                   If provided, all rows with the same group ID are kept in the same fold.

    :return: A list of EstimationValidationIndices tuples containing index sets, one per fold.
    :raises ValueError: If `slices` is less than 2.
    """
    if slices < 2:
        raise ValueError(f'The number of slices is {slices}. It must be at least 2.')

    if groups is None:
        shuffled_data = dataframe.sample(frac=1)
        fold_data = np.array_split(shuffled_data.index, slices)
    else:
        group_ids = dataframe[groups].unique()
        np.random.shuffle(group_ids)
        grouped_ids = np.array_split(group_ids, slices)
        fold_data = [
            dataframe[dataframe[groups].isin(group)].index for group in grouped_ids
        ]

    return [
        EstimationValidationIndices(
            estimation=dataframe.index.difference(fold_data[i]),
            validation=fold_data[i],
        )
        for i in range(slices)
    ]



"""

Nested logit
============

Estimation of a nested logit model using sampling of alternatives.

:author: Michel Bierlaire
:date: Wed Nov  1 18:00:15 2023
"""
import pandas as pd
from biogeme.sampling_of_alternatives import (
    SamplingContext,
    ChoiceSetsGeneration,
    GenerateModel,
    generate_segment_size,
)
from biogeme.expressions import Beta
from biogeme.nests import OneNestForNestedLogit, NestsForNestedLogit
import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from specification import V, combined_variables
from compare import compare
from alternatives import (
    alternatives,
    ID_COLUMN,
    partitions,
    asian,
    all_alternatives,
)

# %%
logger = blog.get_screen_logger(level=blog.INFO)

# %%
SAMPLE_SIZE = 20  # out of 100
SAMPLE_SIZE_MEV = 33  # out of 33
CHOICE_COLUMN = 'nested_0'
PARTITION = 'downtown'
MEV_PARTITION = 'uniform_asian'
MODEL_NAME = f'nested_{PARTITION}_{SAMPLE_SIZE}'
FILE_NAME = f'{MODEL_NAME}.dat'

# %%
the_partition = partitions.get(PARTITION)
if the_partition is None:
    raise ValueError(f'Unknown partition: {PARTITION}')

# %%
segment_sizes = generate_segment_size(SAMPLE_SIZE, the_partition.number_of_segments())

# %%
# We use all alternatives in the nest.
mev_partition = partitions.get(MEV_PARTITION)
if mev_partition is None:
    raise ValueError(f'Unknown partition: {MEV_PARTITION}')
mev_segment_sizes = [SAMPLE_SIZE_MEV]

# %%
observations = pd.read_csv('obs_choice.dat')

# %%
context = SamplingContext(
    the_partition=the_partition,
    sample_sizes=segment_sizes,
    individuals=observations,
    choice_column=CHOICE_COLUMN,
    alternatives=alternatives,
    id_column=ID_COLUMN,
    biogeme_file_name=FILE_NAME,
    utility_function=V,
    combined_variables=combined_variables,
    mev_partition=mev_partition,
    mev_sample_sizes=mev_segment_sizes,
)

# %%
logger.info(context.reporting())

# %%
the_data_generation = ChoiceSetsGeneration(context=context)
the_model_generation = GenerateModel(context=context)

# %%
biogeme_database = the_data_generation.sample_and_merge(recycle=False)

# %%
# Definition of the nest.
mu_asian = Beta('mu_asian', 1.0, 1.0, None, 0)
nest_asian = OneNestForNestedLogit(
    nest_param=mu_asian, list_of_alternatives=asian, name='asian'
)
nests = NestsForNestedLogit(
    choice_set=all_alternatives,
    tuple_of_nests=(nest_asian,),
)

# %%
logprob = the_model_generation.get_nested_logit(nests)

# %%
the_biogeme = bio.BIOGEME(biogeme_database, logprob)
the_biogeme.modelName = MODEL_NAME

# %%
# Calculate the null log likelihood for reporting.
the_biogeme.calculateNullLoglikelihood({i: 1 for i in range(context.total_sample_size)})

# %%
# Estimate the parameters
results = the_biogeme.estimate(recycle=False)

# %%
print(results.short_summary())

# %%
estimated_parameters = results.getEstimatedParameters()
estimated_parameters

# %%
df, msg = compare(estimated_parameters)

# %%
print(df)

# %%
print(msg)



"""Implementation of some multi-objective functions.



:author: Michel Bierlaire
:date: Fri Jul 14 18:35:29 2023

A multi-objective function takes the estimation results, and returns
several indicators. The indicators should be such that the lower, the
better. If an indicator must be maximized, the opposite value should be
returned.

"""

import numpy as np

from biogeme.deprecated import deprecated
from biogeme.results_processing import EstimationResults


def loglikelihood_dimension(results: EstimationResults) -> list[float]:
    """Function returning the negative log likelihood and the number
    of parameters, designed for multi-objective optimization

    :param results: estimation results
    :type results: biogeme.results.bioResults
    """
    if results.raw_estimation_results is None:
        return [float(np.finfo(np.float32).max), float(np.finfo(np.float32).max)]
    return [-results.final_log_likelihood, results.number_of_parameters]


def aic_bic_dimension(results: EstimationResults) -> list[float]:
    """Function returning the AIC, BIC and the number
    of parameters, designed for multi-objective optimization

    :param results: estimation results
    :type results: biogeme.results.bioResults
    """
    if results.raw_estimation_results is None:
        return [
            float(np.finfo(np.float32).max),
            float(np.finfo(np.float32).max),
            float(np.finfo(np.float32).max),
        ]

    return [
        results.akaike_information_criterion,
        results.bayesian_information_criterion,
        results.number_of_parameters,
    ]


@deprecated(new_func=aic_bic_dimension)
def AIC_BIC_dimension(results: EstimationResults) -> list[float]:
    pass



"""Function adding a suffix to variables

Michel Bierlaire
Fri May 02 2025, 17:53:00
"""

from .base_expressions import Expression
from .variable import Variable
from .visitor import ExpressionVisitor

_prefix_suffix_visitor = ExpressionVisitor()
register_prefix_suffix = _prefix_suffix_visitor.register


@register_prefix_suffix(Variable)
def prefix_suffix_variable_name_handler(expr, context):
    expr.name = f"{context['prefix']}{expr.name}{context['suffix']}"
    context['count'] += 1


def add_prefix_suffix_to_all_variables(
    expr: Expression, prefix: str, suffix: str
) -> int:
    context = {'prefix': prefix, 'suffix': suffix, 'count': 0}
    _prefix_suffix_visitor.visit(expr, context)
    return context['count']



import unittest
from unittest.mock import Mock

import pandas as pd

from biogeme.expressions import Beta, Expression, Variable
from biogeme.nests import (
    NestsForCrossNestedLogit,
    NestsForNestedLogit,
    OneNestForCrossNestedLogit,
    OneNestForNestedLogit,
)
from biogeme.sampling_of_alternatives import (
    CrossVariableTuple,
    GenerateModel,
    SamplingContext,
)


class TestGenerateModel(unittest.TestCase):
    def setUp(self):
        # Mocking the context and other required data
        self.mock_context = Mock(spec=SamplingContext)
        self.mock_context.utility_function = Beta('Beta', 0, None, None, 0) * Variable(
            'x'
        )
        self.mock_context.total_sample_size = 5
        self.mock_context.total_mev_sample_size = 4
        self.mock_context.attributes = ['att1', 'att2']
        self.mock_context.mev_prefix = 'mev'
        self.mock_context.mev_partition = None
        self.mock_context.id_column = 'id'

        # Nests for nested logit tests
        a_nest = OneNestForNestedLogit(
            nest_param=Beta('mu', 1, 1, None, 0),
            list_of_alternatives=[1, 5],
            name='any_nest',
        )
        self.nests = NestsForNestedLogit(
            choice_set={1, 2, 3, 4, 5}, tuple_of_nests=(a_nest,)
        )
        first_nest = OneNestForCrossNestedLogit(
            nest_param=Beta('mu_1', 1, 1, None, 0),
            dict_of_alpha={1: 1.0, 3: 0.5},
            name='first_nest',
        )
        second_nest = OneNestForCrossNestedLogit(
            nest_param=Beta('mu_2', 1, 1, None, 0),
            dict_of_alpha={3: 0.5, 5: 1},
            name='second_nest',
        )
        self.cnl_nests = NestsForCrossNestedLogit(
            choice_set={1, 2, 3, 4, 5}, tuple_of_nests=(first_nest, second_nest)
        )

        self.individuals = pd.DataFrame(
            {'choice': [1, 2, 3, 1, 5], 'age': [22, 33, 44, 55, 66]}
        )

        self.alternatives = pd.DataFrame(
            {
                'id': [1, 2, 3, 4, 5],
                'travel_time': [12, 23, 34, 45, 56],
                'travel_cost': [13, 11, 9, 7, 5],
            }
        )

        self.partition = [(1, 2), (3, 4, 5)]
        self.sample_sizes = [1, 2]

        self.utility_function = Beta('beta_cost', 0, None, None, 0) * Variable(
            'travel_cost'
        ) + Beta('beta_age_time', 0, None, None, 0) * Variable('age_time')

        self.combined_variables = [
            CrossVariableTuple('age_time', Variable('age') * Variable('travel_time'))
        ]

    def test_generate_utility(self):
        context = SamplingContext(
            the_partition=self.partition,
            sample_sizes=self.sample_sizes,
            individuals=self.individuals,
            choice_column='choice',
            alternatives=self.alternatives,
            id_column='id',
            biogeme_file_name='tmp_file.dat',
            utility_function=self.utility_function,
            combined_variables=self.combined_variables,
        )
        model = GenerateModel(context)
        utility = model.generate_utility(prefix='prefix_', suffix='_suffix')
        self.assertIsInstance(utility, Expression)
        utility_str = str(utility)
        expected = (
            "((Beta('beta_cost', 0, None, None, "
            "0) * prefix_travel_cost_suffix) + (Beta('beta_age_time', 0, "
            "None, None, 0) * "
            "prefix_age_time_suffix))"
        )
        self.assertEqual(utility_str, expected)

    def test_get_logit(self):
        context = SamplingContext(
            the_partition=self.partition,
            sample_sizes=self.sample_sizes,
            individuals=self.individuals,
            choice_column='choice',
            alternatives=self.alternatives,
            id_column='id',
            biogeme_file_name='tmp_file.dat',
            utility_function=self.utility_function,
            combined_variables=self.combined_variables,
        )
        model = GenerateModel(context)
        logit_expression = model.get_logit()
        self.assertIsInstance(logit_expression, Expression)
        expected_start = 'LogLogit[choice=`0.0`]'
        self.assertTrue(str(logit_expression).startswith(expected_start))

    def test_get_nested_logit(self):
        model = GenerateModel(self.mock_context)
        nested_logit_expression = model.get_nested_logit(self.nests)
        self.assertIsInstance(nested_logit_expression, Expression)
        expected = 'LogLogit[choice=`0.0`]'
        self.assertTrue(str(nested_logit_expression).startswith(expected))

    def test_get_cross_nested_logit(self):
        self.mock_context.cnl_nests = self.cnl_nests
        model = GenerateModel(self.mock_context)
        cross_nested_logit_expression = model.get_cross_nested_logit()
        self.assertIsInstance(cross_nested_logit_expression, Expression)
        expected = 'LogLogit[choice=`0.0`]'
        self.assertTrue(str(cross_nested_logit_expression).startswith(expected))


if __name__ == '__main__':
    unittest.main()



""" Implements the logit model.

:author: Michel Bierlaire
:date: Wed Oct 25 08:43:26 2023
"""

import logging

from biogeme.expressions import (
    Expression,
    LogLogit,
    exp,
    ExpressionOrNumeric,
)

logger = logging.getLogger(__name__)


def loglogit(
    util: dict[int, ExpressionOrNumeric],
    av: dict[int, ExpressionOrNumeric] | None,
    i: ExpressionOrNumeric,
) -> Expression:
    """The logarithm of the logit model

    The model is defined as

    .. math:: \\frac{a_i e^{V_i}}{\\sum_{i=1}^J a_j e^{V_j}}

    :param util: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with util, or
               None. In this case, all alternatives are supposed to be
               always available.

    :param i: id of the alternative for which the probability must be
              calculated.

    :return: choice probability of alternative number i.
    """

    return LogLogit(util, av, i)


def logit(
    util: dict[int, ExpressionOrNumeric],
    av: dict[int, ExpressionOrNumeric] | None,
    i: ExpressionOrNumeric,
) -> Expression:
    """The logit model

    The model is defined as

    .. math:: \\frac{a_i e^{V_i}}{\\sum_{i=1}^J a_j e^{V_j}}

    :param util: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with util, or
               None. In this case, all alternatives are supposed to be
               always available.

    :param i: id of the alternative for which the probability must be
              calculated.

    :return: choice probability of alternative number i.

    """
    return exp(LogLogit(util, av, i))



"""

Base model
==========

Logit model.

:author: Michel Bierlaire, EPFL
:date: Thu Jul 13 16:18:10 2023

"""
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta

# %%
# See :ref:`swissmetro_data`
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

# %%
# Parameters to be estimated.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# %%
# Definition of the utility functions.
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)

# %%
# Create the Biogeme object.
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b00logit'
the_biogeme.generate_html = False
the_biogeme.generate_pickle = False

# %%
# Calculate the null log likelihood for reporting.
the_biogeme.calculateNullLoglikelihood(av)

# %%
# Estimate the parameters
results = the_biogeme.estimate()

# %%
print(results.short_summary())

# %%
# Get the results in a pandas table
pandas_results = results.getEstimatedParameters()
pandas_results
